from multiprocessing import Value
from typing import Callable, List, Iterable, Tuple, Union
from torch import multiprocessing as mp
try:
     mp.set_start_method('spawn')  # needed by CUDA + multiprocessing
except RuntimeError:
    pass
import torch
import gzip
import os
import math
import json
import tqdm
import argparse
import logging
import copy
from . import data_iters
from . import encoder_builders
from pyserini.encode import QueryEncoder, DocumentEncoder

logger = logging.getLogger(__name__)

def one_process(
    encoder_builder: Callable[[int], Union[QueryEncoder, DocumentEncoder]], 
    data_iter: Iterable,
    gpu: int,
    output_dir: str,
    data_range: Tuple[int, int],
    batch_size: int,
    chunk_size: int,
    return_value: Value,
    show_progress_bar: bool
):
    nlines_encoded = 0
    encoder = encoder_builder(gpu)
    batch_ids = []
    batch_texts = []
    term_weights_chunk = []
    begin, end = data_range
    for i, example in enumerate(tqdm.tqdm(data_iter, disable=not show_progress_bar, total=end-begin)):
        if not (begin <= i < end):
            continue
        
        assert type(example) == dict, f'Example must be a dict, but got {type(example)}'
        assert 'id' in example, f'Example must have an id, but got {example}'
        assert 'text' in example, f'Example must have a text, but got {example}'

        batch_ids.append(example['id'])
        batch_texts.append(example['text'])

        if len(batch_ids) >= batch_size or len(batch_ids) + len(term_weights_chunk) >= chunk_size or i == end - 1:
            with torch.no_grad():
                term_weights_batch = encoder.encode(batch_texts)

            term_weights_batch = [
                {
                    "id": doc_id, 
                    "contents": text, 
                    "vector": term_weights
                } for doc_id, text, term_weights in zip(batch_ids, batch_texts, term_weights_batch)
            ]
            term_weights_chunk.extend(term_weights_batch)

            if len(term_weights_chunk) >= chunk_size or i == end - 1:
                chunk_idx = math.ceil((i + 1) / chunk_size) - 1
                fname_save = f'split{chunk_idx}.jsonl.gz'
                logger.info(f'Writing to {fname_save}. i: {i}, range: {data_range}')
                
                os.makedirs(output_dir, exist_ok=True)
                with gzip.open(os.path.join(output_dir, fname_save), 'at+') as f:
                    for line in term_weights_chunk:
                        f.write(json.dumps(line) + '\n')
                        nlines_encoded += 1
                
                term_weights_chunk = []

            batch_ids = []
            batch_texts = []
    
    logger.info(f'Encoded {nlines_encoded} lines. The range is {data_range}')
    return_value.value = nlines_encoded

def _run(
    encoder_builder: Callable[[int], Union[QueryEncoder, DocumentEncoder]],
    data_iter: Iterable, 
    gpus: List[int], 
    output_dir: str,
    batch_size: int, 
    chunk_size: int
):
    nexamples = len(data_iter)
    assert len(set(gpus)) == len(gpus)
    if os.path.exists(output_dir):
        assert len(os.listdir(output_dir)) == 0, 'Please empty the output_dir'

    range_size = nexamples // len(gpus)
    chunk_size = min(chunk_size, range_size-1)  # important for saving all the encoded data !!
    
    ranges = [(b, b+range_size) for b in range(0, nexamples, range_size)]
    ranges[len(gpus)-1] = (ranges[len(gpus)-1][0], ranges[-1][1])  # merge all the leftovers into the last range
    ranges = ranges[:len(gpus)]  # keep only #GPUs ranges

    ranges[-1] = (ranges[-1][0], min(ranges[-1][1], nexamples))  # make sure the last range is right, which is important for the criterion of ending point
    return_values = [Value('d') for _ in range(len(ranges))]
    if len(gpus) > 1:
        ps = [mp.Process(
                target=one_process, 
                args=(
                    encoder_builder, 
                    copy.deepcopy(data_iter),
                    gpu,
                    output_dir,
                    data_range,
                    batch_size,
                    chunk_size,
                    return_values[i],
                    gpu == gpus[0]
                )
            ) for i, (gpu, data_range) in enumerate(zip(gpus, ranges))]
        
        for p in ps:
            p.start()
        for p in ps:
            p.join()
    else:
        # especially for debugging usage
        one_process(
            encoder_builder, 
            data_iter,
            gpus[0],
            output_dir,
            (0, len(data_iter)),
            batch_size,
            chunk_size,
            return_values[0],
            True
        )
    
    nlines_encoded = sum([v.value for v in return_values])
    assert nlines_encoded == nexamples, f'Error: Did not encode all the documents ({nlines_encoded}/{nexamples})!!!'

def run(encoder_name, ckpt_name, data_name, data_dir, gpus, output_dir, batch_size=64, chunk_size=100000):
    encoder_builder = encoder_builders.get_builder(encoder_name, ckpt_name, 'document')
    data_iter = data_iters.build(data_name, data_dir)
    torch.multiprocessing.freeze_support()
    _run(encoder_builder, data_iter, gpus, output_dir, batch_size, chunk_size)
    logger.info(f'{__name__}: Done')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--encoder_name', type=str, choices=["unicoil", "splade", "sparta", "tildev2", "deepimpact"], help="The name of the sparse encoder. For example, splade, unicoil, etc.")
    parser.add_argument('--ckpt_name', type=str, help="The name of the sparse model checkpoint.")
    parser.add_argument('--data_name', type=str, help="The name of the dataset. For example, scifact, trec-covid, etc.")
    parser.add_argument('--data_dir', type=str, help="The directory where the BEIR data is present. For example, if you are using the scifact dataset, this is the directory where the corpus.jsonl is stored.")
    parser.add_argument('--gpus', nargs='+', type=int, help="The gpus to use. For example, --gpus 0 1 2 3")
    parser.add_argument('--output_dir', type=str, help="The output directory where the encoded data will be saved.")
    parser.add_argument('--batch_size', type=int, default=64, help="Default is 64. This is the number of documents to be encoded at a time. If you have a large GPU, you can increase this number to speed up the encoding process.")
    parser.add_argument('--chunk_size', type=int, default=100000, help="Default is 100k. This is the number of documents to be encoded at a time. If you have a large GPU, you can increase this number to speed up the encoding process.")
    args = parser.parse_args()
    run(**vars(args))



#
# Pyserini: Reproducible IR research with sparse and dense representations
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Optional

import torch
if torch.cuda.is_available():
    from torch.cuda.amp import autocast
from transformers import AutoConfig, AutoModel, AutoTokenizer, PreTrainedModel, PretrainedConfig

from pyserini.encode import DocumentEncoder, QueryEncoder


class UniCoilEncoder(PreTrainedModel):
    config_class = AutoConfig
    base_model_prefix = 'coil_encoder'
    load_tf_weights = None

    def __init__(self, config: PretrainedConfig, *model_args, **model_kwargs):
        super().__init__(config, *model_args, **model_kwargs)
        self.config = config
        self.bert = AutoModel.from_config(config)
        self.tok_proj = torch.nn.Linear(config.hidden_size, 1)
        self.init_weights()

    # Copied from transformers.models.bert.modeling_bert.BertPreTrainedModel._init_weights
    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, (torch.nn.Linear, torch.nn.Embedding)):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
        elif isinstance(module, torch.nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, torch.nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def init_weights(self):
        self.bert.init_weights()
        self.tok_proj.apply(self._init_weights)

    def forward(
            self,
            input_ids: torch.Tensor,
            attention_mask: Optional[torch.Tensor] = None,
    ):
        input_shape = input_ids.size()
        device = input_ids.device
        if attention_mask is None:
            attention_mask = (
                torch.ones(input_shape, device=device)
                if input_ids is None
                else (input_ids != self.bert.config.pad_token_id)
            )
        outputs = self.bert(input_ids=input_ids, attention_mask=attention_mask)
        sequence_output = outputs.last_hidden_state
        tok_weights = self.tok_proj(sequence_output)
        tok_weights = torch.relu(tok_weights)
        return tok_weights


class UniCoilDocumentEncoder(DocumentEncoder):
    def __init__(self, model_name, tokenizer_name=None, device='cuda:0'):
        self.device = device
        self.model = UniCoilEncoder.from_pretrained(model_name)
        self.model.to(self.device)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name or model_name)

    def encode(self, texts, titles=None, expands=None, fp16=False):
        if titles:
            texts = [f'{title} {text}' for title, text in zip(titles, texts)]
        max_length = 512  # hardcode for now
        if expands:
            input_ids = self._tokenize_with_injects(texts, expands)
        else:
            input_ids = self.tokenizer(texts, max_length=max_length, padding='longest',
                                       truncation=True, add_special_tokens=True,
                                       return_tensors='pt').to(self.device)["input_ids"]
        if fp16:
            with autocast():
                with torch.no_grad():
                    batch_weights = self.model(input_ids).cpu().detach().numpy()
        else:
            batch_weights = self.model(input_ids).cpu().detach().numpy()
        batch_token_ids = input_ids.cpu().detach().numpy()
        return self._output_to_weight_dicts(batch_token_ids, batch_weights)

    def _output_to_weight_dicts(self, batch_token_ids, batch_weights):
        to_return = []
        for i in range(len(batch_token_ids)):
            weights = batch_weights[i].flatten()
            tokens = self.tokenizer.convert_ids_to_tokens(batch_token_ids[i])
            tok_weights = {}
            for j in range(len(tokens)):
                tok = str(tokens[j])
                weight = float(weights[j])
                if tok == '[CLS]':
                    continue
                if tok == '[PAD]':
                    break
                if tok not in tok_weights:
                    tok_weights[tok] = weight
                elif weight > tok_weights[tok]:
                    tok_weights[tok] = weight
            to_return.append(tok_weights)
        return to_return

    def _tokenize_with_injects(self, texts, expands):
        tokenized = []
        max_len = 0
        for text, expand in zip(texts, expands):
            text_ids = self.tokenizer.encode(text, add_special_tokens=False, max_length=400, truncation=True)
            expand_ids = self.tokenizer.encode(expand, add_special_tokens=False, max_length=100, truncation=True)
            injects = set()
            for tok_id in expand_ids:
                if tok_id not in text_ids:
                    injects.add(tok_id)
            all_tok_ids = [101] + text_ids + [102] + list(injects) + [102]  # 101: CLS, 102: SEP
            tokenized.append(all_tok_ids)
            cur_len = len(all_tok_ids)
            if cur_len > max_len:
                max_len = cur_len
        for i in range(len(tokenized)):
            tokenized[i] += [0] * (max_len - len(tokenized[i]))
        return torch.tensor(tokenized, device=self.device)


class UniCoilQueryEncoder(QueryEncoder):
    def __init__(self, model_name_or_path, tokenizer_name=None, device='cpu'):
        self.device = device
        self.model = UniCoilEncoder.from_pretrained(model_name_or_path)
        self.model.to(self.device)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name or model_name_or_path)

    def encode(self, text, **kwargs):
        max_length = 128  # hardcode for now
        input_ids = self.tokenizer([text], max_length=max_length, padding='longest',
                                   truncation=True, add_special_tokens=True,
                                   return_tensors='pt').to(self.device)["input_ids"]
        batch_weights = self.model(input_ids).cpu().detach().numpy()
        batch_token_ids = input_ids.cpu().detach().numpy()
        return self._output_to_weight_dicts(batch_token_ids, batch_weights)[0]

    def _output_to_weight_dicts(self, batch_token_ids, batch_weights):
        to_return = []
        for i in range(len(batch_token_ids)):
            weights = batch_weights[i].flatten()
            tokens = self.tokenizer.convert_ids_to_tokens(batch_token_ids[i])
            tok_weights = {}
            for j in range(len(tokens)):
                tok = str(tokens[j])
                weight = float(weights[j])
                if tok == '[CLS]':
                    continue
                if tok == '[PAD]':
                    break
                if tok not in tok_weights:
                    tok_weights[tok] = weight
                else:
                    tok_weights[tok] += weight
            to_return.append(tok_weights)
        return to_return



import argparse
from functools import partial
import os
import gzip
import json
from typing import List, Callable
import tqdm
import dill
from multiprocessing import Process, Value
import math
import numpy as np


def one_process(
    fpaths: List[str],
    output_dir: str,
    check_range_only: bool,
    quantize: Callable[[float,], int],
    return_value: Value,
    show_progress_bar: bool
):
    max_value = 0
    for fpath in tqdm.tqdm(fpaths, disable=not show_progress_bar):
        fout = None
        if not check_range_only:
            fpath_output = os.path.join(output_dir, os.path.basename(fpath))
            os.makedirs(output_dir, exist_ok=True)
            fout = gzip.open(fpath_output, 'wt')

        with gzip.open(fpath, 'r') as f:
            for line in f:
                line_dict = json.loads(line)
                term_dict = line_dict['vector']
                if len(term_dict):  # it could be empty!!!
                    max_weight = max(term_dict.values())
                    if max_weight > max_value:
                        max_value = max_weight
                else:
                    print(f'WARNING: Come across empty term weights {line_dict}')

                if not check_range_only:
                    line_dict['vector'] = dict(map(lambda x: (x[0], quantize(x[1])), term_dict.items()))
                    fout.write(json.dumps(line_dict) + '\n')
        
        if fout is not None:
            fout.close()

    return_value.value = max_value

def range_nbits(original_score_range, quantization_nbits, w):
    return int(np.ceil(w/original_score_range * (2**quantization_nbits)))

def ndigits_round(factor, w):
    return round(w * factor)

def build_quantize_fn(method, original_score_range, quantization_nbits, ndigits):
    if method == 'range-nbits':
        assert type(original_score_range) in [float, int] and original_score_range != 0
        assert type(quantization_nbits) is int and quantization_nbits > 0
        quantize_fn = partial(range_nbits, original_score_range, quantization_nbits)
    elif method == 'ndigits-round':
        assert type(ndigits) is int and ndigits > 0
        factor = 10 ** ndigits
        quantize_fn = partial(ndigits_round, factor)
    else:
        raise NotImplementedError
    
    return quantize_fn

def run(
    collection_dir: str,
    output_dir: str,
    method: str,
    original_score_range: float,
    quantization_nbits: int,
    ndigits: int,
    nprocs: int
):
    quantize_fn = None
    
    assert type(output_dir) is str
    quantize_fn = build_quantize_fn(method, original_score_range, quantization_nbits, ndigits)
    print(f'Using quantization method: {method}')

    if original_score_range == -1 and method == 'range-nbits':
        passes = [{'check_range_only': True}, {'check_range_only': False}]
    else:
        passes = [{'check_range_only': False}]

    for i, _pass in enumerate(passes):
        check_range_only = _pass['check_range_only']
        print(f'Pass {i+1}/{len(passes)} for quantization: check_range_only = {check_range_only}')
        if i == 1:
            # Note here we replace `original_score_range` with `max_term_weight``
            quantize_fn = build_quantize_fn(method, max_term_weight, quantization_nbits, ndigits)

        fpaths = [os.path.join(collection_dir, fname) for fname in os.listdir(collection_dir)]
        nprocs = min(len(fpaths), nprocs)
        nfpaths_per_proc = len(fpaths) // nprocs
        ranges = [(b, b+nfpaths_per_proc) for b in range(0, len(fpaths), nfpaths_per_proc)]
        ranges[nprocs-1] = (ranges[nprocs-1][0], ranges[-1][1])  # merge all the leftovers into the last range
        ranges = ranges[:nprocs]  # keep only nprocs ranges
        fpaths_divided = [[fpath for fpath in fpaths[b:e]] for (b, e) in ranges]
        
        return_values = [Value('d') for _ in range(nprocs)]
        processes = [Process(target=one_process, args=(
            fpaths_divided[i],
            output_dir,
            check_range_only,
            quantize_fn,
            return_values[i],
            i == 0
        )) for i in range(nprocs)]    

        [p.start() for p in processes]
        [p.join() for p in processes] 

        max_term_weight = max([v.value for v in return_values])

    print(f'{__name__}: Done')
    print('Max. term weight:', max_term_weight)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--collection_dir')
    parser.add_argument('--output_dir', required=False)
    parser.add_argument('--method', required=False, choices=['range-nbits', 'ndigits-round'])
    # for 'range-nbits':
    parser.add_argument('--original_score_range', type=float, default=5, help='if set to -1, the score range will be computated automatically')
    parser.add_argument('--quantization_nbits', type=int, default=8)
    # for 'ndigits-round':
    parser.add_argument('--ndigits', type=int, default=2, help='2 means *100')
    parser.add_argument('--nprocs', type=int)
    args = parser.parse_args()
    run(**vars(args))



# Adapted from https://github.com/ielab/TILDE/blob/main/modelingv2.py

from collections import defaultdict
import numpy as np
import torch
from transformers import AutoTokenizer, AutoModel, PreTrainedModel, BertConfig, BertModel, PreTrainedTokenizer
from transformers.trainer import Trainer
from torch.utils.data import DataLoader
from typing import Optional
from pyserini.encode import QueryEncoder, DocumentEncoder
import re


def get_stop_ids(tok: PreTrainedTokenizer):
    # hard code for now, from nltk.corpus import stopwords, stopwords.words('english')
    stop_words = set(['i', 'me', 'my', 'myself', 'we', 'our', 'ours', 'ourselves', 'you', "you're", "you've",
                      "you'll", "you'd", 'your', 'yours', 'yourself', 'yourselves', 'he', 'him', 'his',
                      'himself', 'she', "she's", 'her', 'hers', 'herself', 'it', "it's", 'its', 'itself',
                      'they', 'them', 'their', 'theirs', 'themselves', 'what', 'which', 'who', 'whom', 'this',
                      'that', "that'll", 'these', 'those', 'am', 'is', 'are', 'was', 'were', 'be', 'been',
                      'being', 'have', 'has', 'had', 'having', 'do', 'does', 'did', 'doing', 'a', 'an', 'the',
                      'and', 'but', 'if', 'or', 'because', 'as', 'until', 'while', 'of', 'at', 'by', 'for',
                      'with', 'about', 'against', 'between', 'into', 'through', 'during', 'before', 'after',
                      'above', 'below', 'to', 'from', 'up', 'down', 'in', 'out', 'on', 'off', 'over', 'under',
                      'again', 'further', 'then', 'once', 'here', 'there', 'when', 'where', 'why', 'how', 'all',
                      'any', 'both', 'each', 'few', 'more', 'most', 'other', 'some', 'such', 'no', 'nor', 'not',
                      'only', 'own', 'same', 'so', 'than', 'too', 'very', 's', 't', 'can', 'will', 'just', 'don',
                      "don't", 'should', "should've", 'now', 'd', 'll', 'm', 'o', 're', 've', 'y', 'ain', 'aren',
                      "aren't", 'couldn', "couldn't", 'didn', "didn't", 'doesn', "doesn't", 'hadn', "hadn't",
                      'hasn', "hasn't", 'haven', "haven't", 'isn', "isn't", 'ma', 'mightn', "mightn't", 'mustn',
                      "mustn't", 'needn', "needn't", 'shan', "shan't", 'shouldn', "shouldn't", 'wasn', "wasn't",
                      'weren', "weren't", 'won', "won't", 'wouldn', "wouldn't"])
    # keep some common words in ms marco questions
    stop_words.difference_update(["where", "how", "what", "when", "which", "why", "who"])

    vocab = tok.get_vocab()
    tokens = vocab.keys()

    stop_ids = []

    for stop_word in stop_words:
        ids = tok(stop_word, add_special_tokens=False)["input_ids"]
        if len(ids) == 1:
            stop_ids.append(ids[0])

    for token in tokens:
        token_id = vocab[token]
        if token_id in stop_ids:
            continue
        if token == '##s':  # remove 's' suffix
            stop_ids.append(token_id)
        if token[0] == '#' and len(token) > 1:  # skip most of subtokens
            continue
        if not re.match("^[A-Za-z0-9_-]*$", token):  # remove numbers, symbols, etc..
            stop_ids.append(token_id)

    return set(stop_ids)


class TILDEv2(PreTrainedModel):
    config_class = BertConfig
    base_model_prefix = "tildev2"

    def __init__(self, config: BertConfig, train_group_size=8):
        super().__init__(config)
        self.config = config
        # self.bert = BertModel(config)
        self.bert = AutoModel.from_config(config)
        self.tok_proj = torch.nn.Linear(config.hidden_size, 1)
        self.cross_entropy = torch.nn.CrossEntropyLoss(reduction='mean')
        self.train_group_size = train_group_size
        self.init_weights()

    # Copied from transformers.models.bert.modeling_bert.BertPreTrainedModel._init_weights
    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, (torch.nn.Linear, torch.nn.Embedding)):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
        elif isinstance(module, torch.nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, torch.nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def init_weights(self):
        self.bert.init_weights()
        self.tok_proj.apply(self._init_weights)

    def encode(self, **features):
        assert all([x in features for x in ['input_ids', 'attention_mask', 'token_type_ids']])
        model_out = self.bert(**features, return_dict=True)
        reps = self.tok_proj(model_out.last_hidden_state)
        tok_weights = torch.relu(reps)
        return tok_weights

    def forward(self, qry_in, doc_in):
        qry_input = qry_in
        doc_input = doc_in
        doc_out = self.bert(**doc_input, return_dict=True)
        doc_reps = self.tok_proj(doc_out.last_hidden_state)  # D * LD * d


        doc_reps = torch.relu(doc_reps) # relu to make sure no negative weights
        doc_input_ids = doc_input['input_ids']

        # mask ingredients
        qry_input_ids = qry_input['input_ids']
        qry_attention_mask = qry_input['attention_mask']
        self.mask_sep(qry_attention_mask)

        qry_reps = torch.ones_like(qry_input_ids, dtype=torch.float32, device=doc_reps.device).unsqueeze(2)
        tok_scores = self.compute_tok_score_cart(
            doc_reps, doc_input_ids,
            qry_reps, qry_input_ids, qry_attention_mask
        )  # Q * D

        scores = tok_scores

        labels = torch.arange(
            scores.size(0),
            device=doc_input['input_ids'].device,
            dtype=torch.long
        )

        # offset the labels
        labels = labels * self.train_group_size
        loss = self.cross_entropy(scores, labels)
        return loss, scores.view(-1)

    def mask_sep(self, qry_attention_mask):
        sep_pos = qry_attention_mask.sum(1).unsqueeze(1) - 1  # the sep token position
        _zeros = torch.zeros_like(sep_pos)
        qry_attention_mask.scatter_(1, sep_pos.long(), _zeros)
        return qry_attention_mask

    # This function credits to Luyu gao: https://github.com/luyug/COIL/blob/main/modeling.py
    def compute_tok_score_cart(self, doc_reps, doc_input_ids, qry_reps, qry_input_ids, qry_attention_mask):
        qry_input_ids = qry_input_ids.unsqueeze(2).unsqueeze(3)  # Q * LQ * 1 * 1
        doc_input_ids = doc_input_ids.unsqueeze(0).unsqueeze(1)  # 1 * 1 * D * LD
        exact_match = doc_input_ids == qry_input_ids  # Q * LQ * D * LD
        exact_match = exact_match.float()
        scores_no_masking = torch.matmul(
            qry_reps.view(-1, 1),  # (Q * LQ) * d
            doc_reps.view(-1, 1).transpose(0, 1)  # d * (D * LD)
        )
        scores_no_masking = scores_no_masking.view(
            *qry_reps.shape[:2], *doc_reps.shape[:2])  # Q * LQ * D * LD

        scores, _ = (scores_no_masking * exact_match).max(dim=3)  # Q * LQ * D, max pooling
        tok_scores = (scores * qry_attention_mask.unsqueeze(2))[:, 1:].sum(1)

        return tok_scores

class TILDEv2DocumentEncoder(DocumentEncoder):

    def __init__(self, model_name, device) -> None:
        self.model: TILDEv2 = TILDEv2.from_pretrained(model_name)
        self.device = device
        self.model.to(device)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.reverse_voc = {v: k for k, v in self.tokenizer.vocab.items()}
    
    def encode(self, texts, **kwargs):
        p_max_len = kwargs.setdefault('p_max_len', 192)
        inputs = self.tokenizer(
            texts, 
            max_length=p_max_len, 
            padding='longest',
            truncation=True, 
            add_special_tokens=True,
            return_tensors='pt'
        ).to(self.device)
        term_weights = self.model.encode(**inputs).squeeze(-1).cpu().detach().numpy()
        return self._output_to_weight_dicts(inputs['input_ids'], term_weights)

    def _output_to_weight_dicts(self, batch_token_ids, batch_weights):
        to_return = []
        for i in range(len(batch_token_ids)):
            weights = batch_weights[i].flatten()
            tokens = self.tokenizer.convert_ids_to_tokens(batch_token_ids[i])
            tok_weights = {}
            for j in range(len(tokens)):
                tok = str(tokens[j])
                weight = float(weights[j])
                if tok == '[CLS]':
                    continue
                if tok == '[PAD]':
                    break
                if tok not in tok_weights:
                    tok_weights[tok] = weight
                elif weight > tok_weights[tok]:
                    tok_weights[tok] = weight
            to_return.append(tok_weights)
        return to_return


class TILDEv2QueryEncoder(QueryEncoder):

    def __init__(self, model_name_or_path, device='cpu'):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        stop_ids = get_stop_ids(self.tokenizer)
        reverse_voc = {v: k for k, v in self.tokenizer.vocab.items()}
        self.stop_tokens = {reverse_voc[stop_id] for stop_id in stop_ids}

    def encode(self, text, **kwargs):
        tokens = self.tokenizer.tokenize(text)
        tokens = [token for token in tokens if token not in self.stop_tokens]
        term_weights = defaultdict(int)
        for token in tokens:
            term_weights[token] += 1
        return term_weights


# Packaging pyserini.index into both callable and entry point
# With reference to https://github.com/castorini/pyserini/blob/master/pyserini/search/__main__.py
#
# Pyserini: Reproducible IR research with sparse and dense representations
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import argparse
import inspect
import pyserini.index.lucene._base  # important to import this for starting Java services
from jnius import autoclass  # this line should go after


def run(
    collection: str,
    input: str,
    index: str,
    generator: str,
    impact: bool,
    pretokenized: bool,
    threads: int
):
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    assert os.path.isdir(input), ValueError('Argument -input should be a directory.')
    args_strings = []
    for arg in args:
        value = values[arg]
        args_strings.append(f'-{arg}')
        if type(value) is not bool:
            args_strings.append(str(value))
    
    JIndexCollection = autoclass('io.anserini.index.IndexCollection')
    JIndexCollection.main(args_strings)
    print(f'{__name__}: Done')
    

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--collection', '-collection')
    parser.add_argument('--input', '-input')
    parser.add_argument('--index', '-index')
    parser.add_argument('--generator', '-generator')
    parser.add_argument('--impact', '-impact', action='store_true')
    parser.add_argument('--pretokenized', '-pretokenized', action='store_true')
    parser.add_argument('--threads', '-threads')
    args = parser.parse_args()
    run(**vars(args))



import torch
from torch import nn, Tensor
from typing import Iterable, Dict
from .utils import pairwise_dot_score

class FLOPS:
    """constraint from Minimizing FLOPs to Learn Efficient Sparse Representations
    https://arxiv.org/abs/2004.05665
    """
    def __call__(self, batch_rep):
        return torch.sum(torch.mean(torch.abs(batch_rep), dim=0) ** 2)

class MarginMSELossSplade(nn.Module):
    """
    Compute the MSE loss between the |sim(Query, Pos) - sim(Query, Neg)| and |gold_sim(Q, Pos) - gold_sim(Query, Neg)|
    By default, sim() is the dot-product
    For more details, please refer to https://arxiv.org/abs/2010.02666
    """
    def __init__(self, model, similarity_fct = pairwise_dot_score, lambda_d=8e-2, lambda_q=1e-1):
        """
        :param model: SentenceTransformerModel
        :param similarity_fct:  Which similarity function to use
        """
        super(MarginMSELossSplade, self).__init__()
        self.model = model
        self.similarity_fct = similarity_fct
        self.loss_fct = nn.MSELoss()
        self.lambda_d = lambda_d
        self.lambda_q = lambda_q
        self.FLOPS = FLOPS()

    def forward(self, sentence_features: Iterable[Dict[str, Tensor]], labels: Tensor):
        # sentence_features: query, positive passage, negative passage
        reps = [self.model(sentence_feature)['sentence_embedding'] for sentence_feature in sentence_features]
        embeddings_query = reps[0]
        embeddings_pos = reps[1]
        embeddings_neg = reps[2]

        scores_pos = self.similarity_fct(embeddings_query, embeddings_pos)
        scores_neg = self.similarity_fct(embeddings_query, embeddings_neg)
        margin_pred = scores_pos - scores_neg

        flops_doc = self.lambda_d*(self.FLOPS(embeddings_pos) + self.FLOPS(embeddings_neg))
        flops_query = self.lambda_q*(self.FLOPS(embeddings_query))

        return self.loss_fct(margin_pred, labels) + flops_doc + flops_query


import torch
from transformers import PreTrainedModel, AutoModel, AutoConfig

class TILDEv2(PreTrainedModel):
    config_class = AutoConfig
    base_model_prefix = "tildev2"

    def __init__(self, config: AutoConfig, train_group_size=8):
        super().__init__(config)
        self.config = config
        self.model = AutoModel.from_config(config)
        self.tok_proj = torch.nn.Linear(config.hidden_size, 1)
        self.cross_entropy = torch.nn.CrossEntropyLoss(reduction='mean')
        self.train_group_size = train_group_size
        self.init_weights()

    # Copied from transformers.models.bert.modeling_bert.BertPreTrainedModel._init_weights
    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, (torch.nn.Linear, torch.nn.Embedding)):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
        elif isinstance(module, torch.nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, torch.nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def init_weights(self):
        self.model.init_weights()
        self.tok_proj.apply(self._init_weights)

    def encode(self, **features):
        assert all([x in features for x in ['input_ids', 'attention_mask', 'token_type_ids']])
        model_out = self.model(**features, return_dict=True)
        reps = self.tok_proj(model_out.last_hidden_state)
        tok_weights = torch.relu(reps)
        return tok_weights

    def forward(self, qry_in, doc_in):
        qry_input = qry_in
        doc_input = doc_in
        doc_out = self.model(**doc_input, return_dict=True)
        doc_reps = self.tok_proj(doc_out.last_hidden_state)  # D * LD * d


        doc_reps = torch.relu(doc_reps) # relu to make sure no negative weights
        doc_input_ids = doc_input['input_ids']

        # mask ingredients
        qry_input_ids = qry_input['input_ids']
        qry_attention_mask = qry_input['attention_mask']
        self.mask_sep(qry_attention_mask)

        qry_reps = torch.ones_like(qry_input_ids, dtype=torch.float32, device=doc_reps.device).unsqueeze(2)
        tok_scores = self.compute_tok_score_cart(
            doc_reps, doc_input_ids,
            qry_reps, qry_input_ids, qry_attention_mask
        )  # Q * D

        scores = tok_scores

        labels = torch.arange(
            scores.size(0),
            device=doc_input['input_ids'].device,
            dtype=torch.long
        )

        # offset the labels
        labels = labels * self.train_group_size
        loss = self.cross_entropy(scores, labels)
        return loss, scores.view(-1)

    def mask_sep(self, qry_attention_mask):
        sep_pos = qry_attention_mask.sum(1).unsqueeze(1) - 1  # the sep token position
        _zeros = torch.zeros_like(sep_pos)
        qry_attention_mask.scatter_(1, sep_pos.long(), _zeros)
        return qry_attention_mask

    # This function credits to Luyu gao: https://github.com/luyug/COIL/blob/main/modeling.py
    def compute_tok_score_cart(self, doc_reps, doc_input_ids, qry_reps, qry_input_ids, qry_attention_mask):
        qry_input_ids = qry_input_ids.unsqueeze(2).unsqueeze(3)  # Q * LQ * 1 * 1
        doc_input_ids = doc_input_ids.unsqueeze(0).unsqueeze(1)  # 1 * 1 * D * LD
        exact_match = doc_input_ids == qry_input_ids  # Q * LQ * D * LD
        exact_match = exact_match.float()
        scores_no_masking = torch.matmul(
            qry_reps.view(-1, 1),  # (Q * LQ) * d
            doc_reps.view(-1, 1).transpose(0, 1)  # d * (D * LD)
        )
        scores_no_masking = scores_no_masking.view(
            *qry_reps.shape[:2], *doc_reps.shape[:2])  # Q * LQ * D * LD

        scores, _ = (scores_no_masking * exact_match).max(dim=3)  # Q * LQ * D, max pooling
        tok_scores = (scores * qry_attention_mask.unsqueeze(2))[:, 1:].sum(1)

        return tok_scores


import argparse
import csv
import json
import os
from typing import Dict
from .utils import load_qrels
from beir.retrieval.evaluation import EvaluateRetrieval


def load_results(result_path, format) -> Dict[str, Dict[str, float]]:
    results = {}

    # the format is from pyserini's output-format: 
    # https://github.com/castorini/pyserini/blob/c64281a849ae221f96dc80f33085a3ace587250f/pyserini/output_writer.py#L28
    if format == 'msmarco':
        with open(result_path, 'r') as f:
            for line in f:
                qid, doc_id, rank = line.strip().split('\t')
                if str(qid) != str(doc_id): 
                    pseudo_score = 1 / int(rank)
                    results.setdefault(qid, {})
                    results[qid][doc_id] = pseudo_score
    elif format == 'trec':
        with open(result_path, 'r') as f:
            for line in f:
                qid, _, doc_id, rank, score, _ = line.strip().split()
                if str(qid) != str(doc_id): 
                    score = float(score)
                    results.setdefault(qid, {})
                    results[qid][doc_id] = score
    else:
        raise NotImplementedError
    
    return results

def run(result_path, format, qrels_path, output_dir, k_values=[1,3,5,10,100,1000]):
    results = load_results(result_path, format)
    qrels = load_qrels(qrels_path)
    evaluator = EvaluateRetrieval()
    if len(qrels) != len(results):
        missing_queries = set(qrels) - set(results)
        print(f'WARNING: #queries ({len(qrels)}) != |results| ({len(results)}). Queries - results: {set(qrels) - set(results)}')
        for qid in missing_queries:
            results[qid] = {}
            
    # assert len(qrels) == len(results), '#queries should be the same'
    ndcg, _map, recall, precision = evaluator.evaluate(qrels, results, k_values)
    mrr = EvaluateRetrieval.evaluate_custom(qrels, results, k_values, metric='mrr')
    hole = EvaluateRetrieval.evaluate_custom(qrels, results, k_values, metric='hole')

    os.makedirs(output_dir, exist_ok=True)
    with open(os.path.join(output_dir, 'metrics.json'), 'w') as f:
        metrics = {
            'nDCG': ndcg,
            'MAP': _map,
            'Recall': recall,
            'Precision': precision,
            'mrr': mrr,
            'hole': hole,
        }
        json.dump(metrics, f, indent=4)
    print(f'{__name__}: Done')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--result_path')
    parser.add_argument('--format', choices=['msmarco', 'trec'], help='Format of the retrieval result. The formats are from pyserini.output_writer.py')
    parser.add_argument('--qrels_path', help='Path to the BeIR-format file')
    parser.add_argument('--output_dir')
    parser.add_argument('--k_values', nargs='+', type=int, default=[1,2,3,5,10,20,100,1000])
    args = parser.parse_args()
    run(**vars(args))


from json import encoder
from pydoc import Doc
from telnetlib import DO
from numpy import uint
from pyserini.encode import QueryEncoder, DocumentEncoder
from typing import Callable, Union
from .methods import (
    SpladeDocumentEncoder, 
    SpladeQueryEncoder,
    SPARTAQueryEncoder, 
    SPARTADocumentEncoder, 
    TILDEv2QueryEncoder, 
    TILDEv2DocumentEncoder, 
    DeepImpactQueryEncoder, 
    DeepImpactDocumentEncoder,
    UniCoilQueryEncoder,
    UniCoilDocumentEncoder
)
from functools import partial


def unicoil(ckpt_name, etype, device='cpu') -> Union[QueryEncoder, DocumentEncoder]:
    if etype == 'query':
        return UniCoilQueryEncoder(ckpt_name, device=device)        
    elif etype == 'document':
        return UniCoilDocumentEncoder(ckpt_name, device=device)
    else:
        raise ValueError

def splade(ckpt_name, etype, device='cpu'):
    print('WARNING: Since the online checkpoints are only avaible at https://github.com/naver/splade/raw/main/weights/splade_max,'
        'please download them and use a local path for `ckpt_name`')
    if etype == 'query':
        return SpladeQueryEncoder(ckpt_name, device=device)        
    elif etype == 'document':
        return SpladeDocumentEncoder(ckpt_name, device=device)
    else:
        raise ValueError

def sparta(ckpt_name, etype, device='cpu'):
    if etype == 'query':
        return SPARTAQueryEncoder(ckpt_name, device=device)        
    elif etype == 'document':
        return SPARTADocumentEncoder(ckpt_name, device=device)
    else:
        raise ValueError

def tildev2(ckpt_name, etype, device='cpu'):
    if etype == 'query':
        return TILDEv2QueryEncoder(ckpt_name, device=device)        
    elif etype == 'document':
        return TILDEv2DocumentEncoder(ckpt_name, device=device)
    else:
        raise ValueError

def deepimpact(ckpt_name, etype, device='cpu'):
    if etype == 'query':
        return DeepImpactQueryEncoder(ckpt_name, device=device)        
    elif etype == 'document':
        return DeepImpactDocumentEncoder(ckpt_name, device=device)
    else:
        raise ValueError


ENCODER_MAPPING = {
    'unicoil': unicoil,
    'splade': splade,
    'sparta': sparta,
    'tildev2': tildev2,
    'deepimpact': deepimpact
}

def register(encoder_name, builder_fn: Callable[[str, str, Union[str, int]], Union[QueryEncoder, DocumentEncoder]]):
    assert encoder_name not in ENCODER_MAPPING
    ENCODER_MAPPING[encoder_name] = builder_fn

def get_builder(encoder_name, ckpt_name, etype) -> Union[QueryEncoder, DocumentEncoder]:
    assert etype in ['query', 'document']
    
    encoder_name = encoder_name.lower()
    assert encoder_name in ENCODER_MAPPING

    return partial(ENCODER_MAPPING[encoder_name], ckpt_name, etype)


if __name__ == '__main__':
    encoder_builder = get_builder('unicoil', 'castorini/unicoil-noexp-msmarco-passage', 'document')
    encoder: DocumentEncoder = encoder_builder(0)
    text = "Minority interest The reporting of 'minority interest' is a consequence of the requirement by accounting standards to 'fully' consolidate partly owned subsidiaries. Full consolidation, as opposed to partial consolidation, results in financial statements that are constructed as if the parent corporation fully owns these partly owned subsidiaries; except for two line items that reflect partial ownership of subsidiaries: net income to common shareholders and common equity. The two minority interest line items are the net difference between what would have been the common equity and net income to common, if all subsidiaries were fully owned, and the actual ownership of the group. All the other line items in the financial statements assume a fictitious 100% ownership."
    print(encoder.encode([text]))


from collections import defaultdict
import json
import os
from typing import Dict, List
from numpy import sort
from tqdm.auto import tqdm
from . import encoder_builders
from .evaluate import load_results
import argparse

def load_corpus(corpus_path, format) -> Dict[str, str]:
    print('Loading corpus')
    corpus = {}
    format = format.lower()
    if format == 'beir':
        with open(corpus_path, 'r') as f:
            for metadata in tqdm(f):
                line = json.loads(metadata)
                doc_id = line['_id'] if "_id" in line else line['id']
                if "text" in line:
                    text = ' '.join([line['title'], line['text']])
                else:
                    text = ' '.join([line['title'], line['contents']])
                corpus[doc_id] = text
    else:
        raise NotImplementedError
    return corpus

def load_topics(topics_path, format) -> Dict[str, str]:
    format = format.lower()
    queries = {}
    if format == 'beir':
        with open(topics_path, 'r') as f:
            for line in f:
                line_dict = json.loads(line)
                qid = line_dict['_id']
                text = line_dict['text']
                queries[qid] = text
    elif format in ['pyserini', 'anserini']:
        with open(topics_path, 'r') as f:
            # 300674	how many years did william bradford serve as governor of plymouth colony?
            for line in f:
                qid, text = line.strip().split('\t')
                queries[qid] = text
    else:
        raise NotImplementedError
    return queries

def save_results(results, output_dir, format):
    format = format.lower()
    if format == 'trec':
        os.makedirs(output_dir, exist_ok=True)
        with open(os.path.join(output_dir, 'run.tsv'), 'w') as f:
            for qid, rels in results.items():
                # 19335 Q0 7267248 1 1381.000000 Anserini
                for i, (doc_id, score) in enumerate(rels.items()):
                    line = ' '.join([qid, 'Q0', doc_id, str(i + 1), str(score), 'rerank']) + '\n'
                    f.write(line)
    else:
        raise NotImplementedError


def run(
    topics_path,
    topics_format,
    retrieval_result_path,
    retrieval_result_format,
    corpus_path,
    corpus_format,
    encoder_name,
    ckpt_name,
    output_dir,
    output_format,
    device,
    batch_size
):
    queries: Dict[str, str] = load_topics(topics_path, topics_format)
    corpus: Dict[str, str] = load_corpus(corpus_path, corpus_format)
    retrieval_results: Dict[str, Dict[str, float]] = load_results(retrieval_result_path, retrieval_result_format)
    corpus_topk = {doc_id: corpus[doc_id] for _, rels in retrieval_results.items() for doc_id in rels}

    query_encoder_builder = encoder_builders.get_builder(encoder_name, ckpt_name, 'query')
    document_encoder_builder = encoder_builders.get_builder(encoder_name, ckpt_name, 'document')
    query_encoder = query_encoder_builder(device)
    document_encoder = document_encoder_builder(device)

    print('Building lookup table')
    lookup_table = defaultdict(dict)
    doc_ids = list(corpus_topk.keys())
    documents = list(corpus_topk.values())
    for b in tqdm(range(0, len(doc_ids), batch_size)):
        batch = documents[b:b+batch_size]
        term_weight_dicts: List[Dict[str, float]] = document_encoder.encode(batch)  # no quantization here
        for doc_id, term_weights in zip(doc_ids[b:b+batch_size], term_weight_dicts):
            for token, weight in term_weights.items():
                lookup_table[token][doc_id] = weight

    print('Doing re-ranking')
    results_rerank = {}
    for qid, query in tqdm(queries.items()):
        if qid not in retrieval_results:
            print(f'WARNING: Query {qid} is not in the retrieval results. Has ignored it')
            continue

        term_weights = query_encoder.encode(query)
        rels = defaultdict(float)
        for token, weight_query in term_weights.items():
            for doc_id, weight_doc in lookup_table[token].items():
                if doc_id not in retrieval_results[qid]:  # for re-ranking
                    continue
                rels[doc_id] += weight_query * weight_doc
        rels = dict(sorted(rels.items(), key=lambda x:x[1], reverse=True))
        results_rerank[qid] = rels

    print('Saving results')
    save_results(results_rerank, output_dir, output_format)
    print(f'{__name__}: Done')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--topics_path')
    parser.add_argument('--topics_format', default='beir')
    parser.add_argument('--retrieval_result_path')
    parser.add_argument('--retrieval_result_format', default='trec')
    parser.add_argument('--corpus_path')
    parser.add_argument('--corpus_format', default='beir')
    parser.add_argument('--encoder_name')
    parser.add_argument('--ckpt_name')
    parser.add_argument('--output_dir')
    parser.add_argument('--output_format', default='trec')
    parser.add_argument('--device', type=int)
    parser.add_argument('--batch_size', type=int, default=32)
    args = parser.parse_args()
    run(**vars(args))


from builtins import help
from typing import List

from . import encode
from . import quantize
from . import index
from . import reformat_query
from . import search
from . import evaluate
import argparse
import logging
import os


logger = logging.getLogger(__name__)

def run(
    # encode
    encoder_name: str,
    ckpt_name: List[str],
    data_name: str,
    gpus: List[int],
    output_dir: str,

    # reformat_query
    original_query_format: str,
    
    # quantize
    do_quantization: bool,
    quantization_method: str = None,  # TODO: Merge `quantization_method`` and `do_quantization``
    original_score_range: float = None,
    quantization_nbits: int = None,
    ndigits: int = None,
    min_idf: int = -1,

    # search
    topic_split: str = 'test',
    hits: int = 1000,
    output_format_search: str = 'trec',

    # evaluate
    k_values: List[int] = [1,2,3,5,10,20,100,1000],

    # default setting
    batch_size: int = 64,
    chunk_size: int = 100000,
    nprocs: int = 12,
    data_dir: str = None,
    train_data_dir: str = None,
    eval_data_dir: str = None
):
    # 0. Check if the data for training is in the same directory as eval
    # i.e., if you provide data_dir, it means both training and evaluation corpus is present together. 
    if data_dir:
        train_data_dir, eval_data_dir = data_dir, data_dir
    
    # Check if the same checkpoint is used for both doc and query
    if type(ckpt_name) == str:
        query_ckpt, doc_ckpt = ckpt_name, ckpt_name
    elif type(ckpt_name) == list:
        if len(ckpt_name) == 1:
            query_ckpt, doc_ckpt = ckpt_name[0], ckpt_name[0]    
        else:
            query_ckpt, doc_ckpt = ckpt_name[0], ckpt_name[1]
    
    # 1. Encode the documents into term weights
    # The output will be ${output_dir}/collection
    output_dir_encode = os.path.join(output_dir, 'collection')
    if not os.path.exists(output_dir_encode):
        encode.run(encoder_name, doc_ckpt, data_name, train_data_dir, gpus, output_dir_encode, batch_size, chunk_size)
    else:
        logger.info('Escaped encoding due to the existing output file(s)')

    # 2. Quantize the term weights from floats into integers
    # The output will be ${output_dir}-quantized/collection, if `do_quantization == True`
    collection_dir = os.path.join(output_dir, 'collection')
    output_dir_quantize = None
    if do_quantization:
        output_dir += '-quantized'  # TODO: Change this into a specific name
        output_dir_quantize = os.path.join(output_dir, 'collection')
        if not os.path.exists(output_dir_quantize):
            quantize.run(collection_dir, output_dir_quantize, quantization_method, original_score_range, quantization_nbits, ndigits, nprocs)
        else:
            logger.info('Escaped quantization due to the existing output file(s)')
    else:
        logger.info('Escaped quantization due to `do_quantization == False`')

    if do_quantization:
        collection_dir = output_dir_quantize

    # 3. Index the term weights into a Lucene-format index file
    # The output will be ${output_dir}-quantized/index
    output_dir_index = os.path.join(output_dir, 'index')
    if not os.path.exists(output_dir_index):
        index.run('JsonVectorCollection', collection_dir, output_dir_index, 'DefaultLuceneDocumentGenerator', True, True, nprocs)
    else:
        logger.info('Escaped indexing due to the existing output file(s)')

    # 4. Reformat the queries into Pyserini-compatible
    # The output will be under the data directory by default
    if ('beir/' in data_name or 'beir_' in data_name) and eval_data_dir is None:  # TODO: Unify this along with the same snippet within data_iters.py
        eval_data_dir = os.path.join('datasets', data_name.replace('beir_', 'beir/'))
    tsv_queries_path = os.path.join(eval_data_dir, f'queries-{topic_split}.tsv')
    if not os.path.exists(tsv_queries_path):
        reformat_query.run(original_query_format, eval_data_dir, topic_split, None)
    else:
        logger.info('Escaped reformatting quries due to the existing output file(s)')

    # 5. Search the queries over the index
    # The output will be ${output_dir}-quantized/${output_format_search}-format/run.tsv
    output_path_search = os.path.join(output_dir, f'{output_format_search}-format/run.tsv')
    if not os.path.exists(output_path_search):
        search.run(
            topics=tsv_queries_path,
            encoder_name=encoder_name,
            ckpt_name=query_ckpt,
            index=output_dir_index,
            output=output_path_search,
            impact=True,
            hits=hits+1,
            batch_size=batch_size,
            threads=nprocs,
            output_format=output_format_search,
            min_idf=min_idf
        )
    else:
        logger.info('Escaped search due to the existing output file(s)')

    # 6. Do evaluation over the results generated by the last step
    # The output will be under ${output_dir}-quantized/evaluation
    qrels_path = os.path.join(eval_data_dir, 'qrels', f'{topic_split}.tsv')
    output_dir_evaluate = os.path.join(output_dir, 'evaluation')
    if not os.path.exists(output_dir_evaluate):
        evaluate.run(output_path_search, output_format_search, qrels_path, output_dir_evaluate, k_values)
    else:
        logger.info('Escaped evaluation due to the existing output file(s)')

    logger.info(f'{__name__}: Done!')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--encoder_name')
    parser.add_argument('--ckpt_name', nargs='+', type=str, help='Checkpoint name, can be only string or pass query and document checkpoints respectively')
    parser.add_argument('--data_name')
    parser.add_argument('--data_dir', required=False)
    parser.add_argument('--eval_data_dir', required=False)
    parser.add_argument('--train_data_dir', required=False)
    parser.add_argument('--gpus', nargs='+', type=int)
    parser.add_argument('--output_dir')

    parser.add_argument('--min_idf', type=int, default=-1, help='Query tokens with IDF <= this value will be ignored. The default value of -1 means it considers all the tokens')
    parser.add_argument('--do_quantization', action='store_true')
    parser.add_argument('--quantization_method', required=False, choices=['range-nbits', 'ndigits-round'])
    # params for 'range-nbits' quantization technique:
    parser.add_argument('--original_score_range', type=float, default=5)
    parser.add_argument('--quantization_nbits', type=int, default=8)
    # params for 'ndigits-round' quantization technique:
    parser.add_argument('--ndigits', type=int, default=2, help='2 means *100')
    # param: query format whether it is in msmarco or beir format?
    parser.add_argument('--original_query_format', help='e.g. beir')
    # params for evaluation
    parser.add_argument('--topic_split', choices=['train', 'test', 'dev'], help='The queries split for search and evaluation')
    parser.add_argument('--hits', type=int, default=1000)
    parser.add_argument('--output_format_search', type=str, default='trec', choices=['msmarco', 'trec'])

    parser.add_argument('--k_values', nargs='+', type=int, default=[1,2,3,5,10,20,100,1000])

    parser.add_argument('--batch_size', type=int, default=64)
    parser.add_argument('--chunk_size', type=int, default=100000)
    parser.add_argument('--nprocs', type=int, default=12)

    args = parser.parse_args()
    run(**vars(args))




# adapted from https://github.com/castorini/pyserini/blob/master/pyserini/encode/_splade.py

import torch
from transformers import AutoModelForMaskedLM, AutoTokenizer
import numpy as np

from pyserini.encode import QueryEncoder, DocumentEncoder


class SpladeQueryEncoder(QueryEncoder):
    def __init__(self, model_name_or_path, tokenizer_name=None, device='cpu'):
        self.device = device
        self.model = AutoModelForMaskedLM.from_pretrained(model_name_or_path)
        self.model.to(self.device)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name or model_name_or_path)
        self.reverse_voc = {v: k for k, v in self.tokenizer.vocab.items()}

    def encode(self, text, expansion=True, **kwargs):
        max_length = 256  # hardcode for now
        inputs = self.tokenizer([text], max_length=max_length, padding='longest',
                                truncation=True, add_special_tokens=True,
                                return_tensors='pt').to(self.device)
        input_ids = inputs['input_ids']
        input_attention = inputs['attention_mask']
        batch_logits = self.model(input_ids=input_ids, attention_mask=input_attention)['logits']
        token_weights = torch.log(1 + torch.relu(batch_logits)) * input_attention.unsqueeze(-1)
        batch_aggregated_logits, _ = torch.max(token_weights, dim=1)
        batch_aggregated_logits = batch_aggregated_logits.cpu().detach().numpy()

        if expansion:
            return self._output_to_weight_dicts(batch_aggregated_logits)[0]
        else:
            input_ids = input_ids.cpu().detach().numpy()
            return self._output_to_weight_dicts_without_expansion(batch_aggregated_logits, input_ids)[0]

    def _output_to_weight_dicts(self, batch_aggregated_logits):
        to_return = []
        for aggregated_logits in batch_aggregated_logits:
            col = np.nonzero(aggregated_logits)[0]
            weights = aggregated_logits[col]
            d = {self.reverse_voc[k]: round(float(v) * 100) for k, v in zip(list(col), list(weights))}
            to_return.append(d)
        return to_return

    def _output_to_weight_dicts_without_expansion(self, batch_aggregated_logits, input_ids):
        to_return = []
        for idx, aggregated_logits in enumerate(batch_aggregated_logits):
            col = np.nonzero(aggregated_logits)[0]
            col_orig = len(col)
            col = col[np.in1d(col, input_ids[idx])]
            weights = aggregated_logits[col]
            d = {self.reverse_voc[k]: round(float(v) * 100) for k, v in zip(list(col), list(weights))}
            to_return.append(d)
        return to_return


class SpladeDocumentEncoder(DocumentEncoder):
    def __init__(self, model_name_or_path, tokenizer_name=None, device='cpu'):
        self.device = device
        self.model = AutoModelForMaskedLM.from_pretrained(model_name_or_path)
        self.model.to(self.device)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name or model_name_or_path)
        self.reverse_voc = {v: k for k, v in self.tokenizer.vocab.items()}

    def encode(self, texts, expansion=True, **kwargs):
        max_length = 256  # hardcode for now
        inputs = self.tokenizer(texts, max_length=max_length, padding='longest',
                                truncation=True, add_special_tokens=True,
                                return_tensors='pt').to(self.device)
        input_ids = inputs['input_ids']
        input_attention = inputs['attention_mask']
        batch_logits = self.model(input_ids=input_ids, attention_mask=input_attention)['logits']
        token_weights = torch.log(1 + torch.relu(batch_logits)) * input_attention.unsqueeze(-1)
        batch_aggregated_logits, _ = torch.max(token_weights, dim=1)
        batch_aggregated_logits = batch_aggregated_logits.cpu().detach().numpy()

        if expansion:
            return self._output_to_weight_dicts(batch_aggregated_logits)
        else:
            input_ids = input_ids.cpu().detach().numpy()
            return self._output_to_weight_dicts_without_expansion(batch_aggregated_logits, input_ids)

    def _output_to_weight_dicts(self, batch_aggregated_logits):
        to_return = []
        for aggregated_logits in batch_aggregated_logits:
            col = np.nonzero(aggregated_logits)[0]
            weights = aggregated_logits[col]
            d = {self.reverse_voc[k]: float(v) for k, v in zip(list(col), list(weights))}
            to_return.append(d)
        return to_return

    def _output_to_weight_dicts_without_expansion(self, batch_aggregated_logits, input_ids):
        to_return = []
        for idx, aggregated_logits in enumerate(batch_aggregated_logits):
            col = np.nonzero(aggregated_logits)[0]
            col_orig = len(col)
            col = col[np.in1d(col, input_ids[idx])]
            weights = aggregated_logits[col]
            d = {self.reverse_voc[k]: float(v) for k, v in zip(list(col), list(weights))}
            to_return.append(d)
        return to_return



# Packaging pyserini.search into both callable and entry point
# With reference to https://github.com/castorini/pyserini/blob/master/pyserini/search/__main__.py
#
# Pyserini: Reproducible IR research with sparse and dense representations
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import inspect
import os

from tqdm import tqdm
from transformers import AutoTokenizer
from typing import List

from pyserini.analysis import JDefaultEnglishAnalyzer, JWhiteSpaceAnalyzer
from pyserini.output_writer import OutputFormat, get_output_writer
from pyserini.pyclass import autoclass
from pyserini.query_iterator import get_query_iterator, TopicsFormat
from pyserini.search import JDisjunctionMaxQueryGenerator
from pyserini.search.lucene import LuceneImpactSearcher, LuceneSearcher
from pyserini.search.lucene.reranker import ClassifierType, PseudoRelevanceClassifierReranker
from . import encoder_builders


def set_bm25_parameters(searcher, index, k1=None, b=None):
    if k1 is not None or b is not None:
        if k1 is None or b is None:
            print('Must set *both* k1 and b for BM25!')
            exit()
        print(f'Setting BM25 parameters: k1={k1}, b={b}')
        searcher.set_bm25(k1, b)
    else:
        # Automatically set bm25 parameters based on known index:
        if index == 'msmarco-passage' or index == 'msmarco-passage-slim':
            print('MS MARCO passage: setting k1=0.82, b=0.68')
            searcher.set_bm25(0.82, 0.68)
        elif index == 'msmarco-passage-expanded':
            print('MS MARCO passage w/ doc2query-T5 expansion: setting k1=2.18, b=0.86')
            searcher.set_bm25(2.18, 0.86)
        elif index == 'msmarco-doc' or index == 'msmarco-doc-slim':
            print('MS MARCO doc: setting k1=4.46, b=0.82')
            searcher.set_bm25(4.46, 0.82)
        elif index == 'msmarco-doc-per-passage' or index == 'msmarco-doc-per-passage-slim':
            print('MS MARCO doc, per passage: setting k1=2.16, b=0.61')
            searcher.set_bm25(2.16, 0.61)
        elif index == 'msmarco-doc-expanded-per-doc':
            print('MS MARCO doc w/ doc2query-T5 (per doc) expansion: setting k1=4.68, b=0.87')
            searcher.set_bm25(4.68, 0.87)
        elif index == 'msmarco-doc-expanded-per-passage':
            print('MS MARCO doc w/ doc2query-T5 (per passage) expansion: setting k1=2.56, b=0.59')
            searcher.set_bm25(2.56, 0.59)


class AttrDict(dict):
  def __init__(self, *args, **kwargs):
      super(AttrDict, self).__init__(*args, **kwargs)
      self.__dict__ = self


def run(
    topics: str,
    index: str,
    output: str,
    topics_format: str = TopicsFormat.DEFAULT.value,
    output_format: str = OutputFormat.TREC.value,
    max_passage: bool = False,
    max_passage_hits: int = 100,
    max_passage_delimiter: str = '#',
    batch_size: int = 1,
    threads: int = 1,
    remove_duplicates: bool = False,
    hits: int = 1000,
    impact: bool = False,
    encoder_name: str = None,
    ckpt_name: str = None,
    tokenizer: str = None,
    min_idf: int = -1,
    bm25: bool = False,
    rm3: bool = False,
    qld: bool = False,
    language: str = 'en',
    prcl: List[ClassifierType] = [],
    k1: float = None,
    b: float = None,
    vectorizer: str = None,
    fields = None,
    stopwords: str = None,
    r: int = 10,
    n: int = 100,
    alpha: float = 0.5,
    dismax: bool = False,
    tiebreaker: float = 0.0
):
    frame = inspect.currentframe()
    args, _, _, values = inspect.getargvalues(frame)
    args = AttrDict(dict(zip(args, map(lambda arg: values[arg], args))))

    query_iterator = get_query_iterator(args.topics, TopicsFormat(args.topics_format))
    topics = query_iterator.topics

    if not args.impact:
        if os.path.exists(args.index):
            # create searcher from index directory
            searcher = LuceneSearcher(args.index)
        else:
            # create searcher from prebuilt index name
            searcher = LuceneSearcher.from_prebuilt_index(args.index)
    elif args.impact:
        ######## build query encoder by encoder name and checkpoint name ##########
        encoder_builder = encoder_builders.get_builder(encoder_name, ckpt_name, 'query')
        args.encoder = encoder_builder()  # By default this will use CPU
        ###########################################################################
        if os.path.exists(args.index):
            searcher = LuceneImpactSearcher(args.index, args.encoder, args.min_idf)
        else:
            searcher = LuceneImpactSearcher.from_prebuilt_index(args.index, args.encoder, args.min_idf)

    if args.language != 'en':
        searcher.set_language(args.language)

    if not searcher:
        exit()

    search_rankers = []

    if args.qld:
        search_rankers.append('qld')
        searcher.set_qld()
    elif args.bm25:
        search_rankers.append('bm25')
        set_bm25_parameters(searcher, args.index, args.k1, args.b)

    if args.rm3:
        search_rankers.append('rm3')
        searcher.set_rm3()

    fields = dict()
    if args.fields:
        fields = dict([pair.split('=') for pair in args.fields])
        print(f'Searching over fields: {fields}')

    query_generator = None
    if args.dismax:
        query_generator = JDisjunctionMaxQueryGenerator(args.tiebreaker)
        print(f'Using dismax query generator with tiebreaker={args.tiebreaker}')
    
    if args.tokenizer != None:
        analyzer = JWhiteSpaceAnalyzer()
        searcher.set_analyzer(analyzer)
        print(f'Using whitespace analyzer because of pretokenized topics')
        tokenizer = AutoTokenizer.from_pretrained(args.tokenizer)
        print(f'Using {args.tokenizer} to preprocess topics')

    if args.stopwords:
        analyzer = JDefaultEnglishAnalyzer.fromArguments('porter', False, args.stopwords)
        searcher.set_analyzer(analyzer)
        print(f'Using custom stopwords={args.stopwords}')

    # get re-ranker
    use_prcl = args.prcl and len(args.prcl) > 0 and args.alpha > 0
    if use_prcl is True:
        ranker = PseudoRelevanceClassifierReranker(
            searcher.index_dir, args.vectorizer, args.prcl, r=args.r, n=args.n, alpha=args.alpha)

    # build output path
    output_path = args.output
    if output_path is None:
        if use_prcl is True:
            clf_rankers = []
            for t in args.prcl:
                if t == ClassifierType.LR:
                    clf_rankers.append('lr')
                elif t == ClassifierType.SVM:
                    clf_rankers.append('svm')

            r_str = f'prcl.r_{args.r}'
            n_str = f'prcl.n_{args.n}'
            a_str = f'prcl.alpha_{args.alpha}'
            clf_str = 'prcl_' + '+'.join(clf_rankers)
            tokens1 = ['run', args.topics, '+'.join(search_rankers)]
            tokens2 = [args.vectorizer, clf_str, r_str, n_str, a_str]
            output_path = '.'.join(tokens1) + '-' + '-'.join(tokens2) + ".txt"
        else:
            tokens = ['run', args.topics, '+'.join(search_rankers), 'txt']
            output_path = '.'.join(tokens)

    print(f'Running {args.topics} topics, saving to {output_path}...')
    tag = output_path[:-4] if args.output is None else 'Anserini'

    output_writer = get_output_writer(output_path, OutputFormat(args.output_format), 'w',
                                      max_hits=args.hits, tag=tag, topics=topics,
                                      use_max_passage=args.max_passage,
                                      max_passage_delimiter=args.max_passage_delimiter,
                                      max_passage_hits=args.max_passage_hits)

    with output_writer:
        batch_topics = list()
        batch_topic_ids = list()
        for index, (topic_id, text) in enumerate(tqdm(query_iterator, total=len(topics.keys()))):
            if (args.tokenizer != None):
                toks = tokenizer.tokenize(text)
                text = ' '
                text = text.join(toks)
            if args.batch_size <= 1 and args.threads <= 1:
                if args.impact:
                    hits = searcher.search(text, args.hits, fields=fields)
                else:
                    hits = searcher.search(text, args.hits, query_generator=query_generator, fields=fields)
                results = [(topic_id, hits)]
            else:
                batch_topic_ids.append(str(topic_id))
                batch_topics.append(text)
                if (index + 1) % args.batch_size == 0 or \
                        index == len(topics.keys()) - 1:
                    if args.impact:
                        results = searcher.batch_search(
                            batch_topics, batch_topic_ids, args.hits, args.threads, fields=fields
                        )
                    else:
                        results = searcher.batch_search(
                            batch_topics, batch_topic_ids, args.hits, args.threads,
                            query_generator=query_generator, fields=fields
                        )
                    results = [(id_, results[id_]) for id_ in batch_topic_ids]
                    batch_topic_ids.clear()
                    batch_topics.clear()
                else:
                    continue

            for topic, hits in results:
                # do rerank
                if use_prcl and len(hits) > (args.r + args.n):
                    docids = [hit.docid.strip() for hit in hits]
                    scores = [hit.score for hit in hits]
                    scores, docids = ranker.rerank(docids, scores)
                    docid_score_map = dict(zip(docids, scores))
                    for hit in hits:
                        hit.score = docid_score_map[hit.docid.strip()]
                        
                if args.remove_duplicates:
                    seen_docids = set()
                    dedup_hits = []
                    for hit in hits:
                        if hit.docid.strip() in seen_docids:
                            continue
                        seen_docids.add(hit.docid.strip())
                        dedup_hits.append(hit)
                    hits = dedup_hits

                # write results
                output_writer.write(topic, hits)

            results.clear()
    print(f'{__name__}: Done')

def define_search_args(parser):
    parser.add_argument('--index', type=str, metavar='path to index or index name', required=True,
                        help="Path to Lucene index or name of prebuilt index.")

    parser.add_argument('--impact', action='store_true', help="Use Impact.")
    # parser.add_argument('--encoder', type=str, default=None, help="encoder name")
    parser.add_argument('--encoder_name', type=str, default=None, help="encoder name")
    parser.add_argument('--ckpt_name', type=str, default=None, help="encoder ckpt name or path")
    parser.add_argument('--min-idf', type=int, default=-1, help='Query tokens with IDF <= this value will be ignored. The default value of -1 means it considers all the tokens')

    parser.add_argument('--bm25', action='store_true', default=True, help="Use BM25 (default).")
    parser.add_argument('--k1', type=float, help='BM25 k1 parameter.')
    parser.add_argument('--b', type=float, help='BM25 b parameter.')

    parser.add_argument('--rm3', action='store_true', help="Use RM3")
    parser.add_argument('--qld', action='store_true', help="Use QLD")

    parser.add_argument('--language', type=str, help='language code for BM25, e.g. zh for Chinese', default='en')

    parser.add_argument('--prcl', type=ClassifierType, nargs='+', default=[],
                        help='Specify the classifier PseudoRelevanceClassifierReranker uses.')
    parser.add_argument('--prcl.vectorizer', dest='vectorizer', type=str,
                        help='Type of vectorizer. Available: TfidfVectorizer, BM25Vectorizer.')
    parser.add_argument('--prcl.r', dest='r', type=int, default=10,
                        help='Number of positive labels in pseudo relevance feedback.')
    parser.add_argument('--prcl.n', dest='n', type=int, default=100,
                        help='Number of negative labels in pseudo relevance feedback.')
    parser.add_argument('--prcl.alpha', dest='alpha', type=float, default=0.5,
                        help='Alpha value for interpolation in pseudo relevance feedback.')

    parser.add_argument('--fields', metavar="key=value", nargs='+',
                        help='Fields to search with assigned float weights.')
    parser.add_argument('--dismax', action='store_true', default=False,
                        help='Use disjunction max queries when searching multiple fields.')
    parser.add_argument('--dismax.tiebreaker', dest='tiebreaker', type=float, default=0.0,
                        help='The tiebreaker weight to use in disjunction max queries.')

    parser.add_argument('--stopwords', type=str, help='Path to file with customstopwords.')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Search a Lucene index.')
    define_search_args(parser)
    parser.add_argument('--topics', type=str, metavar='topic_name', required=True,
                        help="Name of topics. Available: robust04, robust05, core17, core18.")
    parser.add_argument('--hits', type=int, metavar='num',
                        required=False, default=1000, help="Number of hits.")
    parser.add_argument('--topics-format', type=str, metavar='format', default=TopicsFormat.DEFAULT.value,
                        help=f"Format of topics. Available: {[x.value for x in list(TopicsFormat)]}")
    parser.add_argument('--output-format', type=str, metavar='format', default=OutputFormat.TREC.value,
                        help=f"Format of output. Available: {[x.value for x in list(OutputFormat)]}")
    parser.add_argument('--output', type=str, metavar='path',
                        help="Path to output file.")
    parser.add_argument('--max-passage',  action='store_true',
                        default=False, help="Select only max passage from document.")
    parser.add_argument('--max-passage-hits', type=int, metavar='num', required=False, default=100,
                        help="Final number of hits when selecting only max passage.")
    parser.add_argument('--max-passage-delimiter', type=str, metavar='str', required=False, default='#',
                        help="Delimiter between docid and passage id.")
    parser.add_argument('--batch-size', type=int, metavar='num', required=False,
                        default=1, help="Specify batch size to search the collection concurrently.")
    parser.add_argument('--threads', type=int, metavar='num', required=False,
                        default=1, help="Maximum number of threads to use.")
    parser.add_argument('--tokenizer', type=str, help='tokenizer used to preprocess topics')
    parser.add_argument('--remove-duplicates', action='store_true', default=False, help="Remove duplicate docs.")
    args = parser.parse_args()
    run(**vars(args))


import torch
import torch.nn as nn
from itertools import accumulate
from transformers import AutoModel, AutoTokenizer
from .utils import clean_query, clean_passage

class DeepImpact(nn.Module):
    def __init__(self, model_name, device, max_seq_length=300):
        super().__init__()
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.bert = AutoModel.from_pretrained(model_name)
        self.config = self.bert.config
        hidden_dropout_prob = self.config.hidden_dropout_prob if "hidden_dropout_prob" in dict(self.config) else \
            self.config.attention_dropout
        self.device = device
        self.max_seq_length = max_seq_length
        self.impact_score_encoder = nn.Sequential(
            nn.Linear(768, 768),
            nn.ReLU(),
            nn.Dropout(hidden_dropout_prob),
            nn.Linear(768, 1),
            nn.ReLU()
        )
        self.init_weights()

    def convert_example(self, passage, max_seq_length):
        max_length = min(self.max_seq_length, max_seq_length)
        return self.tokenizer.encode_plus(passage, add_special_tokens=True, max_length=max_length, truncation=True, padding="max_length")

    def tokenize(self, query, passage):
        query_tokens = list(set(clean_query(query).strip().split()))
        content = clean_passage(passage).strip()
        doc_tokens = content.split()

        # NOTE: The following line accounts for CLS!
        tokenized = self.tokenizer.tokenize(content)
        word_indexes = list(accumulate([-1] + tokenized, lambda a, b: a + int(not b.startswith('##'))))
        match_indexes = list(set([doc_tokens.index(t) for t in query_tokens if t in doc_tokens]))
        term_indexes = [word_indexes.index(idx) for idx in match_indexes]

        a = [idx for i, idx in enumerate(match_indexes) if term_indexes[i] < self.max_seq_length]
        b = [idx for idx in term_indexes if idx < self.max_seq_length]

        return content, tokenized, a, b, len(word_indexes) + 2

    def forward(self, queries, passages):
        bsize = len(queries)
        pairs = []
        X, pfx_sum, pfx_sumX = [], [], []
        total_size, total_sizeX, max_seq_length = 0, 0, 0

        doc_partials = []
        pre_pairs = []

        for query, passage in zip(queries, passages):
            tokens, tokenized, term_idxs, token_idxs, seq_length = self.tokenize(query, passage)
            max_seq_length = max(max_seq_length, seq_length)

            pfx_sumX.append(total_sizeX)
            total_sizeX += len(term_idxs)

            tokens_split = tokens.split()

            doc_partials.append([(total_size + idx, tokens_split[i]) for idx, i in enumerate(term_idxs)])
            total_size += len(doc_partials[-1])
            pfx_sum.append(total_size)

            pre_pairs.append((tokenized, token_idxs))

        for tokenized, token_idxs in pre_pairs:
            pairs.append(self.convert_example(tokenized, max_seq_length))
            X.append(token_idxs)

        input_ids = torch.tensor([f['input_ids'] for f in pairs], dtype=torch.long).to(self.device)
        attention_mask = torch.tensor([f['attention_mask'] for f in pairs], dtype=torch.long).to(self.device)
        token_type_ids = torch.tensor([f['token_type_ids'] for f in pairs], dtype=torch.long).to(self.device)

        outputs = self.bert.forward(input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)

        hidden_state = outputs[0]

        def one(i):
            if len(X[i]) > 0:
                l = [hidden_state[i, j] for j in X[i]]  # + [mismatch_scores[i, j] for j in all_mismatches[i]]
                return torch.stack(l)
            return torch.tensor([]).to(self.device)

        pooled_output = torch.cat([one(i) for i in range(bsize)])

        bsize = len(pooled_output)

        if bsize == 0:
            term_scores = []
            for doc in doc_partials:
                term_scores.append([])
                for (idx, term) in doc:
                    term_scores[-1].append((term, 0.0))

            return torch.tensor([[0.0]] * len(Q)).to(self.device), term_scores

        y_score = self.impact_score_encoder(pooled_output)

        x = torch.arange(bsize).expand(len(pfx_sum), bsize) < torch.tensor(pfx_sum).unsqueeze(1)
        y = torch.arange(bsize).expand(len(pfx_sum), bsize) >= torch.tensor([0] + pfx_sum[:-1]).unsqueeze(1)
        mask = (x & y).to(self.device)

        y_scorex = list(y_score.cpu())
        term_scores = []
        for doc in doc_partials:
            term_scores.append([])
            for (idx, term) in doc:
                term_scores[-1].append((term, y_scorex[idx]))

        return (mask.type(torch.float32) @ y_score), term_scores #, ordered_terms #, num_exceeding_fifth

    def encode(self, D, max_seq_length):
        if max_seq_length % 10 == 0:
            print("#>>>   max_seq_length = ", max_seq_length)

        bsize = len(D)
        offset = 0
        pairs, X = [], []

        for tokenized_content, terms in D:
            terms = [(t, idx, offset + pos) for pos, (t, idx) in enumerate(terms)]
            offset += len(terms)
            pairs.append(self.convert_example(tokenized_content, max_seq_length))
            X.append(terms)

        input_ids = torch.tensor([f['input_ids'] for f in pairs], dtype=torch.long).to(self.device)
        attention_mask = torch.tensor([f['attention_mask'] for f in pairs], dtype=torch.long).to(self.device)
        token_type_ids = torch.tensor([f['token_type_ids'] for f in pairs], dtype=torch.long).to(self.device)

        outputs = self.bert.forward(input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)

        hidden_state = outputs[0]
        pooled_output = torch.cat([hidden_state[i, list(map(lambda x: x[1], X[i]))] for i in range(bsize)])

        y_score = self.impact_score_encoder(pooled_output)
        y_score = y_score.squeeze().cpu().numpy().tolist()
        term_scores = [[(term, y_score[pos]) for term, _, pos in terms] for terms in X]

        return term_scores


import random
import re
import datasets
from typing import Union, List

from torch.utils.data import Dataset
from transformers import PreTrainedTokenizer
from nltk.corpus import stopwords

class GroupedMarcoTrainDataset(Dataset):
    query_columns = ['qid', 'query']
    document_columns = ['pid', 'passage']

    def __init__(
            self,
            path_to_tsv: Union[List[str], str],
            tokenizer: PreTrainedTokenizer,
            q_max_len: int,
            p_max_len: int,
            train_group_size: int,
            cache_dir: str,
            stopwords: bool = False,
    ):
        self.tok = tokenizer
        self.SEP = [self.tok.sep_token_id]
        self.stop_ids = self.get_stop_ids() if stopwords else set()
        self.q_max_len = q_max_len
        self.p_max_len = p_max_len
        self.train_group_size = train_group_size

        self.nlp_dataset = datasets.load_dataset(
            'json',
            cache_dir=cache_dir,
            data_files=path_to_tsv,
            ignore_verifications=False,
            features=datasets.Features({
                'qry': {
                    'qid': datasets.Value('string'),
                    'query': [datasets.Value('int32')],
                },
                'pos': [{
                    'pid': datasets.Value('string'),
                    'passage': [datasets.Value('int32')],
                }],
                'neg': [{
                    'pid': datasets.Value('string'),
                    'passage': [datasets.Value('int32')],
                }]}
            )
        )['train']
        self.total_len = len(self.nlp_dataset)

    def create_one_example(self, text_encoding, is_query=False):
        item = self.tok.encode_plus(
            text_encoding,
            truncation='only_first',
            return_attention_mask=False,
            max_length=self.q_max_len if is_query else self.p_max_len,
        )
        return item

    def __len__(self):
        return self.total_len

    def get_stop_ids(self):
        stop_words = set(stopwords.words('english'))
        # keep some common words in ms marco questions
        stop_words.difference_update(["where", "how", "what", "when", "which", "why", "who"])

        vocab = self.tok.get_vocab()
        tokens = vocab.keys()

        stop_ids = []

        for stop_word in stop_words:
            ids = self.tok(stop_word, add_special_tokens=False)["input_ids"]
            if len(ids) == 1:
                # bad_token.append(stop_word)
                stop_ids.append(ids[0])

        for token in tokens:
            token_id = vocab[token]
            if token_id in stop_ids:
                continue
            if token == '##s':
                stop_ids.append(token_id)
            if token[0] == '#' and len(token) > 1:
                continue
            if not re.match("^[A-Za-z0-9_-]*$", token):
                # bad_token.append(token)
                stop_ids.append(token_id)

        return set(stop_ids)

    def __getitem__(self, item):
        group = self.nlp_dataset[item]
        group_batch = []
        qid, qry = (group['qry'][k] for k in self.query_columns)
        qry = [id for id in qry if id not in self.stop_ids]

        encoded_query = self.create_one_example(qry, is_query=True)
        _, pos_psg = [
            random.choice(group['pos'])[k] for k in self.document_columns]
        group_batch.append(self.create_one_example(pos_psg))
        if len(group['neg']) < self.train_group_size - 1:
            negs = random.choices(group['neg'], k=self.train_group_size - 1)
        else:
            negs = random.sample(group['neg'], k=self.train_group_size - 1)
        for neg_entry in negs:
            _, neg_psg = [neg_entry[k] for k in self.document_columns]
            group_batch.append(self.create_one_example(neg_psg))

        return encoded_query, group_batch



import csv
import json
from typing import Dict

def load_qrels(qrels_path) -> Dict[str, Dict[str, float]]:
    # adapted from BeIR: 
    # https://github.com/beir-cellar/beir/blob/main/beir/datasets/data_loader.py#L114
    reader = csv.reader(open(qrels_path, encoding="utf-8"), delimiter="\t", quoting=csv.QUOTE_MINIMAL)
    next(reader)  # skip header
    
    qrels = {}
    for row in reader:
        query_id, corpus_id, score = row[0], row[1], int(row[2])
        
        if query_id not in qrels:
            qrels[query_id] = {corpus_id: score}
        else:
            qrels[query_id][corpus_id] = score
    
    return qrels


def load_queries(queries_path) -> Dict[str, str]:
    # adapted from BEIR:
    # https://github.com/beir-cellar/beir/blob/main/beir/datasets/data_loader.py#L107
    queries = {}
    with open(queries_path, encoding='utf8') as fIn:
        for line in fIn:
            line = json.loads(line)
            queries[line.get("_id")] = line.get("text")
    
    return queries


# Adapted from https://github.com/DI4IR/SIGIR2021/blob/main/src/model.py

import os
from typing import Optional, Union
import torch
import torch.nn as nn

from random import sample, shuffle, randint

from itertools import accumulate
from transformers import AutoConfig, AutoTokenizer, AutoModel, PreTrainedModel
from pyserini.encode import QueryEncoder, DocumentEncoder
import re

MAX_LENGTH = 300

STOPLIST = ["a", "about", "also", "am", "an", "and", "another", "any", "anyone", "are", "aren't", "as", "at", "be",
            "been", "being", "but", "by", "despite", "did", "didn't", "do", "does", "doesn't", "doing", "done", "don't",
            "each", "etc", "every", "everyone", "for", "from", "further", "had", "hadn't", "has", "hasn't", "have",
            "haven't", "having", "he", "he'd", "he'll", "her", "here", "here's", "hers", "herself", "he's",
            "him", "himself", "his", "however", "i", "i'd", "if", "i'll", "i'm", "in", "into", "is", "isn't", "it",
            "its", "it's", "itself", "i've", "just", "let's", "like", "lot", "may", "me", "might", "mightn't",
            "my", "myself", "no", "nor", "not", "of", "on", "onto", "or", "other", "ought", "oughtn't", "our", "ours",
            "ourselves", "out", "over", "shall", "shan't", "she", "she'd", "she'll", "she's", "since", "so", "some",
            "something", "such", "than", "that", "that's", "the", "their", "theirs", "them", "themselves", "then",
            "there", "there's", "these", "they", "they'd", "they'll", "they're", "they've", "this", "those", "through",
            "tht", "to", "too", "usually", "very", "via", "was", "wasn't", "we", "we'd", "well", "we'll", "were",
            "we're", "weren't", "we've", "will", "with", "without", "won't", "would", "wouldn't", "yes", "yet", "you",
            "you'd", "you'll", "your", "you're", "yours", "yourself", "yourselves", "you've"]

printable = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ ')
printableX = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-. ')
printable3X = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ,.- ')

printableD = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.- ')
printable3D = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.- ')

STOPLIST_ = list(map(lambda s: ''.join(filter(lambda x: x in printable, s)), STOPLIST))

STOPLIST = {}
for w in STOPLIST_:
    STOPLIST[w] = True

def cleanD(s, join=True):
    s = [(x.lower() if x in printable3X else ' ') for x in s]
    s = [(x if x in printableX else ' ' + x + ' ') for x in s]
    s = ''.join(s).split()
    s = [(w if '.' not in w else (' . ' if len(max(w.split('.'), key=len)) > 1 else '').join(w.split('.'))) for w in s]
    s = ' '.join(s).split()
    s = [(w if '-' not in w else w.replace('-', '') + ' ( ' + ' '.join(w.split('-')) + ' ) ') for w in s]
    s = ' '.join(s).split()
    # s = [w for w in s if w not in STOPLIST]

    return ' '.join(s) if join else s


def cleanQ(s, join=True):
    s = [(x.lower() if x in printable3D else ' ') for x in s]
    s = [(x if x in printableD else ' ' + x + ' ') for x in s]
    s = ''.join(s).split()
    s = [(w if '.' not in w else (' ' if len(max(w.split('.'), key=len)) > 1 else '').join(w.split('.'))) for w in s]
    s = ' '.join(s).split()
    s = [(w if '-' not in w else (' ' if len(min(w.split('-'), key=len)) > 1 else '').join(w.split('-'))) for w in s]
    s = ' '.join(s).split()
    s = [w for w in s if w not in STOPLIST]

    return ' '.join(s) if join else s


class DeepImpact(PreTrainedModel):
    config_class = AutoConfig
    # base_model_prefix = 'coil_encoder'
    base_model_prefix = 'deep_impact'

    def __init__(self, config, *model_args, **model_kwargs):
        super(DeepImpact, self).__init__(config)
        # self.tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
        # self.bert = BertModel(config)
        # self.tokenizer = AutoTokenizer.from_config(config)
        self.bert = AutoModel.from_config(config)
        self.pre_classifier = nn.Linear(config.hidden_size, config.hidden_size)
        self.classifier = nn.Linear(config.hidden_size, 1)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        # self.impact_score_encoder = nn.Sequential(
        #     nn.Linear(config.hidden_size, config.hidden_size),
        #     nn.ReLU(),
        #     nn.Dropout(config.hidden_dropout_prob),
        #     nn.Linear(config.hidden_size, 1),
        #     nn.ReLU()
        # )
        self.init_weights()
                
    def impact_score_encoder(self, hs):
        for module in [self.pre_classifier, self.relu, self.dropout, self.classifier, self.relu]:
            hs = module(hs)
        return hs

    def convert_example(self, d, max_seq_length):
        max_length = min(MAX_LENGTH, max_seq_length)
        inputs = self.tokenizer.encode_plus(d, add_special_tokens=True, max_length=max_length, truncation=True)

        padding_length = max_length - len(inputs["input_ids"])
        attention_mask = ([1] * len(inputs["input_ids"])) + ([0] * padding_length)
        input_ids = inputs["input_ids"] + ([0] * padding_length)
        token_type_ids = inputs["token_type_ids"] + ([0] * padding_length)

        return {'input_ids': input_ids, 'attention_mask': attention_mask, 'token_type_ids': token_type_ids}

    def tokenize(self, q, d):
        query_tokens = list(set(cleanQ(q).strip().split()))  # [:10]

        content = cleanD(d).strip()
        doc_tokens = content.split()

        # NOTE: The following line accounts for CLS!
        tokenized = self.tokenizer.tokenize(content)
        word_indexes = list(accumulate([-1] + tokenized, lambda a, b: a + int(not b.startswith('##'))))
        match_indexes = list(set([doc_tokens.index(t) for t in query_tokens if t in doc_tokens]))
        term_indexes = [word_indexes.index(idx) for idx in match_indexes]

        a = [idx for i, idx in enumerate(match_indexes) if term_indexes[i] < MAX_LENGTH]
        b = [idx for idx in term_indexes if idx < MAX_LENGTH]

        return content, tokenized, a, b, len(word_indexes) + 2

    def tokenize_documents(self, d):
        d = cleanD(d, join=False)
        content = ' '.join(d)
        tokenized_content = self.tokenizer.tokenize(content)

        terms = list(set([(t, d.index(t)) for t in d]))  # Quadratic!
        word_indexes = list(accumulate([-1] + tokenized_content, lambda a, b: a + int(not b.startswith('##'))))
        terms = [(t, word_indexes.index(idx)) for t, idx in terms]
        terms = [(t, idx) for (t, idx) in terms if idx < MAX_LENGTH]

        return tokenized_content, terms

    def forward(self, Q, D):
        bsize = len(Q)
        pairs = []
        X, pfx_sum, pfx_sumX = [], [], []
        total_size, total_sizeX, max_seq_length = 0, 0, 0

        doc_partials = []
        pre_pairs = []

        for q, d in zip(Q, D):
            tokens, tokenized, term_idxs, token_idxs, seq_length = self.tokenize(q, d)
            max_seq_length = max(max_seq_length, seq_length)

            pfx_sumX.append(total_sizeX)
            total_sizeX += len(term_idxs)

            tokens_split = tokens.split()

            doc_partials.append([(total_size + idx, tokens_split[i]) for idx, i in enumerate(term_idxs)])
            total_size += len(doc_partials[-1])
            pfx_sum.append(total_size)

            pre_pairs.append((tokenized, token_idxs))

        for tokenized, token_idxs in pre_pairs:
            pairs.append(self.convert_example(tokenized, max_seq_length))
            X.append(token_idxs)

        input_ids = torch.tensor([f['input_ids'] for f in pairs], dtype=torch.long).to(self.device)
        attention_mask = torch.tensor([f['attention_mask'] for f in pairs], dtype=torch.long).to(self.device)
        token_type_ids = torch.tensor([f['token_type_ids'] for f in pairs], dtype=torch.long).to(self.device)

        outputs = self.bert.forward(input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)

        hidden_state = outputs[0]

        def one(i):
            if len(X[i]) > 0:
                l = [hidden_state[i, j] for j in X[i]]  # + [mismatch_scores[i, j] for j in all_mismatches[i]]
                return torch.stack(l)
            return torch.tensor([]).to(self.device)

        pooled_output = torch.cat([one(i) for i in range(bsize)])

        bsize = len(pooled_output)

        if bsize == 0:
            term_scores = []
            for doc in doc_partials:
                term_scores.append([])
                for (idx, term) in doc:
                    term_scores[-1].append((term, 0.0))

            return torch.tensor([[0.0]] * len(Q)).to(self.device), term_scores

        y_score = self.impact_score_encoder(pooled_output)

        x = torch.arange(bsize).expand(len(pfx_sum), bsize) < torch.tensor(pfx_sum).unsqueeze(1)
        y = torch.arange(bsize).expand(len(pfx_sum), bsize) >= torch.tensor([0] + pfx_sum[:-1]).unsqueeze(1)
        mask = (x & y).to(self.device)

        y_scorex = list(y_score.cpu())
        term_scores = []
        for doc in doc_partials:
            term_scores.append([])
            for (idx, term) in doc:
                term_scores[-1].append((term, y_scorex[idx]))

        return (mask.type(torch.float32) @ y_score), term_scores #, ordered_terms #, num_exceeding_fifth

    def index(self, D, max_seq_length):
        bsize = len(D)
        offset = 0
        pairs, X = [], []

        for tokenized_content, terms in D:
            terms = [(t, idx, offset + pos) for pos, (t, idx) in enumerate(terms)]
            offset += len(terms)
            pairs.append(self.convert_example(tokenized_content, max_seq_length))
            X.append(terms)

        input_ids = torch.tensor([f['input_ids'] for f in pairs], dtype=torch.long).to(self.device)
        attention_mask = torch.tensor([f['attention_mask'] for f in pairs], dtype=torch.long).to(self.device)
        token_type_ids = torch.tensor([f['token_type_ids'] for f in pairs], dtype=torch.long).to(self.device)

        outputs = self.bert.forward(input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)

        hidden_state = outputs[0]
        pooled_output = torch.cat([hidden_state[i, list(map(lambda x: x[1], X[i]))] for i in range(bsize)])

        y_score = self.impact_score_encoder(pooled_output)
        y_score = y_score.squeeze().cpu().numpy().tolist()
        # term_scores = [[(term, y_score[pos]) for term, _, pos in terms] for terms in X]
        # term_scores = [{term: round(y_score[pos], 3) for term, _, pos in terms} for terms in X]
        term_scores = [{term: y_score[pos] for term, _, pos in terms} for terms in X]
        return term_scores


class DeepImpactDocumentEncoder(DocumentEncoder):

    def __init__(self, model_name, device) -> None:
        self.model: DeepImpact = DeepImpact.from_pretrained(model_name)
        self.device = device
        self.model.to(device)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=False)
        self.model.tokenizer = self.tokenizer
        self.reverse_voc = {v: k for k, v in self.tokenizer.vocab.items()}

    def encode(self, texts, **kwargs):
        super_batch = list(map(self.model.tokenize_documents, texts))
        max_seq_length = max([len(tokenized_content) for tokenized_content, terms in super_batch]) + 2
        return self.model.index(super_batch, max_seq_length)

class DeepImpactQueryEncoder(QueryEncoder):

    def __init__(self, model_name_or_path, device='cpu'):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, use_fast=False)

    def encode(self, text, **kwargs):
        query_tokens = list(set(cleanQ(text).strip().split()))  # [:10]
        return dict(zip(query_tokens, [1] * len(query_tokens)))


if __name__ == '__main__':
    query_encoder = DeepImpactQueryEncoder('deepimpact-bert-base')
    doc_encoder = DeepImpactDocumentEncoder('deepimpact-bert-base', 'cuda')

    q = 'What is the presence of communication communication communication amid scientific minds?'
    d = 'The presence of communication amid scientific minds was equally important to the success of the Manhattan Project as scientific intellect was. The only cloud hanging over the impressive achievement of the atomic researchers and engineers is what their success truly meant; hundreds of thousands of innocent lives obliterated. [SEP] how be world are results public an have which it want result significant in were they communicate on amongst involved affect believe power impact successful importance scientists bomb contribute outcome why know effect so who effects about did for purpose valuable following most accomplishment a communications'

    with torch.no_grad():
        tw_q = query_encoder.encode(q)
        tw_d = doc_encoder.encode([d])[0]
    
    print(tw_q)
    print(tw_d)
    tw_shared = {t: tw_d[t] for t in tw_q if t in tw_d}
    print(tw_shared)
    print(sum(tw_shared.values()))



import torch
from torch import nn, Tensor
from typing import Iterable, Dict
from .utils import dot_score

class FLOPS:
    """constraint from Minimizing FLOPs to Learn Efficient Sparse Representations
    https://arxiv.org/abs/2004.05665
    """
    def __call__(self, batch_rep):
        return torch.sum(torch.mean(torch.abs(batch_rep), dim=0) ** 2)

class MultipleNegativesRankingLossSplade(nn.Module):
    """
        This loss expects as input a batch consisting of sentence pairs (a_1, p_1), (a_2, p_2), ..., (a_n, p_n)
        where we assume that (a_i, p_i) are positive pairs and (a_i, p_j) for i!=j negative pairs.
        For each a_i, it uses all other p_j as negative samples, i.e., for a_i, we have 1 positive example (p_i) and
        n-1 negative examples (p_j). It then minimizes the negative log-likehood for softmax normalized scores.
        This loss function works great to train embeddings for retrieval setups where you have positive pairs (e.g. (query, relevant_doc))
        as it will sample in each batch n-1 negative docs randomly. The performance usually increases with increasing batch sizes.
        For more information, see: https://arxiv.org/pdf/1705.00652.pdf
        (Efficient Natural Language Response Suggestion for Smart Reply, Section 4.4)
        You can also provide one or multiple hard negatives per anchor-positive pair by structering the data like this:
        (a_1, p_1, n_1), (a_2, p_2, n_2)
        Here, n_1 is a hard negative for (a_1, p_1). The loss will use for the pair (a_i, p_i) all p_j (j!=i) and all n_j as negatives.
    """
    def __init__(self, model, scale: float = 1.0, similarity_fct = dot_score, lambda_d: float = 0.0008, lambda_q: float = 0.0006):
        """
        :param model: SentenceTransformer model
        :param scale: Output of similarity function is multiplied by scale value
        :param similarity_fct: similarity function between sentence embeddings. By default, cos_sim. Can also be set to dot product (and then set scale to 1)
        """
        super(MultipleNegativesRankingLossSplade, self).__init__()
        self.model = model
        self.scale = scale
        self.similarity_fct = similarity_fct
        self.cross_entropy_loss = nn.CrossEntropyLoss()
        self.lambda_d = lambda_d
        self.lambda_q = lambda_q
        self.FLOPS = FLOPS()

    def forward(self, sentence_features: Iterable[Dict[str, Tensor]], labels: Tensor):
        reps = [self.model(sentence_feature)['sentence_embedding'] for sentence_feature in sentence_features]
        embeddings_a = reps[0]
        embeddings_b = torch.cat(reps[1:])

        scores = self.similarity_fct(embeddings_a, embeddings_b) * self.scale
        labels = torch.tensor(range(len(scores)), dtype=torch.long, device=scores.device)  # Example a[i] should match with b[i]

        flops_doc = self.lambda_d*(self.FLOPS(embeddings_b))
        flops_query = self.lambda_q*(self.FLOPS(embeddings_a))

        return self.cross_entropy_loss(scores, labels) + flops_doc + flops_query

    def get_config_dict(self):
        return {'scale': self.scale, 'similarity_fct': self.similarity_fct.__name__, "lambda_q": self.lambda_q, "lambda_d": self.lambda_d}


import torch
from torch import Tensor

def pairwise_dot_score(a: Tensor, b: Tensor):
    """
   Computes the pairwise dot-product dot_prod(a[i], b[i])
   :return: Vector with res[i] = dot_prod(a[i], b[i])
   """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    return (a * b).sum(dim=-1)

def dot_score(a: Tensor, b: Tensor):
    """
    Computes the dot-product dot_prod(a[i], b[j]) for all i and j
    :return: Matrix with res[i][j]  = dot_prod(a[i], b[j])
    """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    if len(a.shape) == 1:
        a = a.unsqueeze(0)

    if len(b.shape) == 1:
        b = b.unsqueeze(0)

    return torch.mm(a, b.transpose(0, 1))


from .splade import SpladeDocumentEncoder, SpladeQueryEncoder
from .sparta import SPARTADocumentEncoder, SPARTAQueryEncoder
from .tildev2 import TILDEv2DocumentEncoder, TILDEv2QueryEncoder
from .deepimpact import DeepImpactDocumentEncoder, DeepImpactQueryEncoder
from .unicoil import UniCoilDocumentEncoder, UniCoilQueryEncoder


import os, sys
from typing import Dict, List, Tuple, Optional, Any, Union

from transformers.trainer import Trainer, nested_detach
from transformers.trainer_utils import PredictionOutput, EvalPrediction

import torch
from torch import nn
from torch.cuda.amp import autocast
from torch.utils.data import DataLoader
import logging

logger = logging.getLogger(__name__)

class SPARSETrainer(Trainer):
    def _save(self, output_dir: Optional[str] = None):
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.info("Saving model checkpoint to %s", output_dir)
        
        self.model.save_pretrained(output_dir)
        if self.tokenizer is not None and self.is_world_process_zero():
            self.tokenizer.save_pretrained(output_dir)
        torch.save(self.args, os.path.join(output_dir, "training_args.bin"))

    def _prepare_inputs(self, inputs):
        prepared = {}
        for k, v in inputs.items():
            prepared[k] = {}
            for sk, sv in v.items():
                if isinstance(sv, torch.Tensor):
                    prepared[k][sk] = sv.to(self.args.device)

        return prepared

    def create_optimizer_and_scheduler(self, num_training_steps: int):
        if self.args.warmup_ratio > 0:
            self.args.warmup_steps = num_training_steps * self.args.warmup_ratio

        super().create_optimizer_and_scheduler(num_training_steps)

    def get_train_dataloader(self) -> DataLoader:
        if self.train_dataset is None:
            raise ValueError("Trainer: training requires a train_dataset.")
        train_sampler = self._get_train_sampler()

        return DataLoader(
            self.train_dataset,
            batch_size=self.args.train_batch_size,
            sampler=train_sampler,
            collate_fn=self.data_collator,
            drop_last=True,
            num_workers=self.args.dataloader_num_workers,
        )

    def prediction_step(
            self,
            model: nn.Module,
            inputs: Tuple[Dict[str, Union[torch.Tensor, Any]]],
            prediction_loss_only: bool,
            ignore_keys: Optional[List[str]] = None,
    ) -> Tuple[Optional[float], Optional[torch.Tensor], Optional[torch.Tensor]]:

        qry, doc = self._prepare_inputs(inputs)[:2]
        if ignore_keys is None:
            raise NotImplementedError

        with torch.no_grad():
            if self.args.fp16:
                with autocast():
                    outputs = model(qry, doc)
            else:
                outputs = model(qry, doc)

            loss = outputs.loss
            logits = outputs.scores

        if prediction_loss_only:
            return (loss, None, None)

        logits = nested_detach(logits)
        if len(logits) == 1:
            logits = logits[0]

        labels = None

        return (loss, logits, labels)

    def prediction_loop(
            self,
            *args,
            **kwargs
    ) -> PredictionOutput:
        pred_outs = super().prediction_loop(*args, **kwargs)
        preds, label_ids, metrics = pred_outs.predictions, pred_outs.label_ids, pred_outs.metrics
        preds = preds.squeeze()
        if self.compute_metrics is not None:
            metrics_no_label = self.compute_metrics(EvalPrediction(predictions=preds, label_ids=label_ids))
        else:
            metrics_no_label = {}

        for key in list(metrics_no_label.keys()):
            if not key.startswith("eval_"):
                metrics_no_label[f"eval_{key}"] = metrics_no_label.pop(key)
        return PredictionOutput(predictions=preds, label_ids=label_ids, metrics={**metrics, **metrics_no_label})


# Adapted from https://github.com/nreimers/beir-sparta/blob/main/SPARTA.py
# and https://github.com/nreimers/beir-sparta/blob/main/eval_msmarco.py

from collections import defaultdict
import torch
from transformers import AutoTokenizer, AutoModel
from pyserini.encode import QueryEncoder, DocumentEncoder


class SPARTADocumentEncoder(torch.nn.Module, DocumentEncoder):
    def __init__(self, model_name, device): #SpanBERT/spanbert-base-cased'): #bert-base-uncased    #distilbert-base-uncased #distilroberta-base
        super().__init__()
        print("Model name:", model_name)
        self.bert_model = AutoModel.from_pretrained(model_name)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.bert_model.to(device)
        self.score_bias = torch.nn.Parameter(torch.tensor([1.0], device=device))
        self.device = device
        self.max_length = 300
        #####
        self.bert_input_emb = self.bert_model.embeddings.word_embeddings(torch.tensor(list(range(0, len(self.tokenizer))), device=device))  # for building term weights
        self.reverse_voc = {v: k for k, v in self.tokenizer.vocab.items()}
        self.special_token_embedding_to_zero = False  # used during inference

    def bert_embeddings(self, input_ids):
        return self.bert_model.embeddings.word_embeddings(input_ids)

    def query_embeddings(self, query):
        queries_batch = self.tokenizer(query, padding=True, truncation=True, return_tensors='pt', add_special_tokens=False, max_length=self.max_length).to(self.device)
        queries_embeddings = self.bert_embeddings(queries_batch['input_ids'])
        return queries_embeddings

    def passage_embeddings(self, passages):
        passage_batch = self.tokenizer(passages, padding=True, truncation=True, return_tensors='pt', max_length=self.max_length).to(self.device)
        passage_embeddings = self.bert_model(**passage_batch).last_hidden_state
        return passage_embeddings

    def compute_scores(self, query_embeddings, passage_embeddings):
        ### Eq. 4 - Term matching
        scores = []
        for idx in range(len(query_embeddings)):        #TODO: use correct pytorch function for this
            scores.append(torch.matmul(query_embeddings[idx], passage_embeddings.transpose(1, 2)))
        scores = torch.stack(scores)
        #print("Scores:", scores.shape)
        max_scores = torch.max(scores, dim=-1).values
        #print("Max-Scores:", max_scores.shape)

        ### Eq. 5 - ReLu
        relu_scores = torch.relu(max_scores)    #torch.relu(max_scores + self.score_bias)  #Bias score does not change that much?
        #print("ReLu-Scores:", relu_scores.shape)

        ### Eq. 6 - Final Score
        final_scores = torch.sum(torch.log(relu_scores + 1), dim=-1) #.unsqueeze(dim=0)
        #print("Final scores:", final_scores.shape)
        return final_scores

    def forward(self, queries, passages):
        query_embeddings = self.query_embeddings(queries)
        passage_embeddings = self.passage_embeddings(passages)
        return self.compute_scores(query_embeddings, passage_embeddings)
    
    ###
    def _set_special_token_embedding_to_zero(self):
        if self.bert_model.training == True:
            return

        if self.special_token_embedding_to_zero:
            return
        
        for special_id in self.tokenizer.all_special_ids:
            self.bert_input_emb[special_id] = 0 * self.bert_input_emb[special_id]
        
        self.special_token_embedding_to_zero = True
            
    ###
    def encode(self, texts, **kwargs):
        self._set_special_token_embedding_to_zero()  # Important for full reproduction (although it seems to have little influence on the performance)
        
        term_weights_batch = []
        sparse_vec_size = kwargs.setdefault('sparse_vec_size', 2000)  # TODO: Make this into the search.py cli arguments
        assert sparse_vec_size <= len(self.tokenizer)

        tokens = self.tokenizer(texts, padding=True, truncation=True, return_tensors='pt', max_length=500).to(self.device)
        passage_embeddings = self.bert_model(**tokens).last_hidden_state
        for passage_emb in passage_embeddings:  # TODO: Optimize this by batch operations
            scores = torch.matmul(self.bert_input_emb, passage_emb.transpose(0, 1))
            max_scores = torch.max(scores, dim=-1).values
            relu_scores = torch.relu(max_scores) #Eq. 5
            final_scores = torch.log(relu_scores + 1)  # Eq. 6, final score

            top_results = torch.topk(final_scores, k=sparse_vec_size)
            tids = top_results[1].cpu().detach().tolist()
            scores = top_results[0].cpu().detach().tolist()
            
            term_weights = {}
            for tid, score in zip(tids, scores):
                if score > 0:
                    term_weights[self.reverse_voc[tid]] = score
                else:
                    break

            term_weights_batch.append(term_weights)
        
        return term_weights_batch

class SPARTAQueryEncoder(QueryEncoder):

    def __init__(self, model_name_or_path, device='cpu'):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        self.reverse_voc = {v: k for k, v in self.tokenizer.vocab.items()}

    def encode(self, text, **kwargs):
        token_ids = self.tokenizer(text, add_special_tokens=False)['input_ids']
        tokens = [self.reverse_voc[token_id] for token_id in token_ids]
        term_weights = defaultdict(int)
        
        # Important for reproducing the results:
        # Note that in Pyserini/Anserini, the query term weights are maintained by JHashMap,
        # which will keep only one term weight for identical terms
        for token in tokens:
            term_weights[token] += 1
        return term_weights


import argparse
import os
from .utils import load_qrels, load_queries

def convert_beir_queries(data_dir, topic_split, output_dir=None):
    if output_dir is None:
        output_dir = data_dir
    else:
        os.makedirs(output_dir, exist_ok=True)
    
    queries = load_queries(os.path.join(data_dir, "queries.jsonl"))
    qrels = load_qrels(os.path.join(data_dir, 'qrels', f'{topic_split}.tsv'))
    queries = {qid: queries[qid] for qid in qrels}
    
    with open(os.path.join(output_dir, f'queries-{topic_split}.tsv'), 'w') as fOut:
        for query_id, query in queries.items():
            query = query.replace('\t', '')
            line = '\t'.join([str(query_id), query]) + '\n'
            fOut.write(line)

def run(original_format, data_dir, topic_split, output_dir=None):
    original_format = original_format.lower()

    if original_format == 'beir':
        convert_beir_queries(data_dir, topic_split, output_dir)
    else:
        raise NotImplementedError
    
    print(f'{__name__}: Done')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--original_format')
    parser.add_argument('--data_dir')
    parser.add_argument('--topic_split')
    parser.add_argument('--output_dir', default=None)
    args = parser.parse_args()
    run(**vars(args))



from sentence_transformers import InputExample
from torch.utils.data import Dataset
import random

class MSMARCODataset(Dataset):
    def __init__(self, queries, corpus):
        self.queries = queries
        self.queries_ids = list(queries.keys())
        self.corpus = corpus

        for qid in self.queries:
            self.queries[qid]['pos'] = list(self.queries[qid]['pos'])
            self.queries[qid]['hard_neg'] = list(self.queries[qid]['hard_neg'])
            random.shuffle(self.queries[qid]['hard_neg'])

    def __getitem__(self, item):
        query = self.queries[self.queries_ids[item]]
        query_text = query['query']

        pos_id = query['pos'].pop(0)    #Pop positive and add at end
        pos_text = self.corpus[pos_id]["text"]
        query['pos'].append(pos_id)
        pos_score = float(query['pos_scores'][pos_id])

        neg_id = query['hard_neg'].pop(0)    #Pop negative and add at end
        neg_text = self.corpus[neg_id]["text"]
        query['hard_neg'].append(neg_id)
        neg_score = float(query['hard_neg_scores'][neg_id])

        return InputExample(texts=[query_text, pos_text, neg_text], label=(pos_score - neg_score))

    def __len__(self):
        return len(self.queries)


from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    readme = fh.read()

setup(
    name="sprint-toolkit",
    version="0.0.3",
    author="Nandan Thakur",
    author_email="nandant@gmail.com",
    description="SPRINT: A Unified Toolkit for Evaluating and Demystifying Zero-shot Neural Sparse Retrieval",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/thakur-nandan/sprint",
    project_urls={
        "Bug Tracker": "https://github.com/thakur-nandan/sprint/issues",
    },
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Intended Audience :: Science/Research",
    ],
    python_requires=">=3.7",
    install_requires=[
        "beir",
        "pyserini"
    ],
    keywords="Information Retrieval Toolkit Sparse Retrievers Networks BERT PyTorch IR NLP deep learning"
)


from torch.utils.data import Dataset
from beir import util
import os
import linecache
import json

class BeIRDataIterator(Dataset):

    def __init__(self, data_dir: str):
        self.data_dir = data_dir
        self.nexamples = len(linecache.getlines(data_dir))
        self.sep = ' '
    
    def __getitem__(self, index: int):
        if index >= self.nexamples:
            raise StopIteration

        line = linecache.getline(self.data_dir, index + 1)
        line_dict = json.loads(line)
        example = {}
        example['id'] = line_dict['_id']
        example['text'] = self.sep.join([line_dict['title'], line_dict['text']])
        return example

    def __len__(self):
        return self.nexamples


def beir(dataset: str, data_dir: str = None):
    if data_dir is None:
        assert dataset is not None
        url = f"https://public.ukp.informatik.tu-darmstadt.de/thakur/BEIR/datasets/{dataset}.zip"
        out_dir = 'datasets/beir'
        data_dir = os.path.join(out_dir, dataset)
        if not os.path.exists(data_dir):
            util.download_and_unzip(url, out_dir)

    return BeIRDataIterator(os.path.join(data_dir, 'corpus.jsonl'))


def build(data_name: str, data_dir: str):
    data_name = data_name.lower()
    
    if 'beir' in data_name:
        data_name = data_name.replace('beir/', '')
        data_name = data_name.replace('beir_', '')
        return beir(data_name, data_dir)
    else:
        raise NotImplementedError


from transformers import AutoModel, PreTrainedModel, TrainingArguments
from transformers.modeling_outputs import BaseModelOutputWithPooling

import torch
from torch import nn, Tensor
import torch.distributed as dist

import os
import logging
from typing import Dict
from torch.cuda.amp import autocast

logger = logging.getLogger(__name__)


class UniCOIL(nn.Module):
    def __init__(self, model: PreTrainedModel, model_args, data_args,
                 train_args: TrainingArguments):
        super().__init__()
        self.model: PreTrainedModel = model
        self.cross_entropy = nn.CrossEntropyLoss(reduction='mean')
        self.data_args, self.model_args, self.train_args = data_args, model_args, train_args
        self.tok_proj = nn.Linear(768, model_args.token_dim)
        self.cls_proj = nn.Linear(768, model_args.cls_dim)

        if model_args.token_norm_after:
            self.ln_tok = nn.LayerNorm(model_args.token_dim)
        if model_args.cls_norm_after:
            self.ln_cls = nn.LayerNorm(model_args.cls_dim)

    @classmethod
    def from_pretrained(
            cls, model_args, data_args, train_args: TrainingArguments,
            *args, **kwargs
    ):
        hf_model = AutoModel.from_pretrained(*args, **kwargs)
        model = UniCOIL(hf_model, model_args, data_args, train_args)
        path = args[0]
        if os.path.exists(os.path.join(path, 'pytorch_model.bin')):
            logger.info('loading extra weights from local files')
            model_dict = torch.load(os.path.join(path, 'pytorch_model.bin'), map_location="cpu")
            load_result = model.load_state_dict(model_dict, strict=False)
        return model

    def save_pretrained(self, output_dir: str):
        self.model.save_pretrained(output_dir)
        model_dict = self.state_dict()
        
        encoder_state = {}
        for key, value in model_dict.items():
            if key.startswith('model'):
                encoder_state[key.replace("model", "coil_encoder.bert")] = value
            elif key.startswith('tok_proj'):
                encoder_state["coil_encoder." + key] = value
        
        torch.save(encoder_state, os.path.join(output_dir, 'pytorch_model.bin'))

        # for k in hf_weight_keys:
        #     model_dict.pop(k)
        # torch.save(model_dict, os.path.join(output_dir, 'model.pt'))
        # torch.save([self.data_args, self.model_args, self.train_args], os.path.join(output_dir, 'args.pt'))

    def encode(self, **features):
        assert all([x in features for x in ['input_ids', 'attention_mask', 'token_type_ids']])
        model_out: BaseModelOutputWithPooling = self.model(**features, return_dict=True)
        cls = self.cls_proj(model_out.last_hidden_state[:, 0])
        reps = self.tok_proj(model_out.last_hidden_state)
        if self.model_args.cls_norm_after:
            cls = self.ln_cls(cls)
        if self.model_args.token_norm_after:
            reps = self.ln_tok(reps)

        if self.model_args.token_rep_relu:
            reps = torch.relu(reps)

        return cls, reps

    def forward(self, qry_input: Dict, doc_input: Dict):
        qry_out: BaseModelOutputWithPooling = self.model(**qry_input, return_dict=True)
        doc_out: BaseModelOutputWithPooling = self.model(**doc_input, return_dict=True)

        qry_cls = self.cls_proj(qry_out.last_hidden_state[:, 0])
        doc_cls = self.cls_proj(doc_out.last_hidden_state[:, 0])

        qry_reps = self.tok_proj(qry_out.last_hidden_state)  # Q * LQ * d
        doc_reps = self.tok_proj(doc_out.last_hidden_state)  # D * LD * d

        if self.model_args.cls_norm_after:
            qry_cls, doc_cls = self.ln_cls(qry_cls), self.ln_cls(doc_cls)
        if self.model_args.token_norm_after:
            qry_reps, doc_reps = self.ln_tok(qry_reps), self.ln_tok(doc_reps)

        if self.model_args.token_rep_relu:
            qry_reps = torch.relu(qry_reps)
            doc_reps = torch.relu(doc_reps)

        # mask ingredients
        doc_input_ids: Tensor = doc_input['input_ids']
        qry_input_ids: Tensor = qry_input['input_ids']
        qry_attention_mask: Tensor = qry_input['attention_mask']

        self.mask_sep(qry_attention_mask)

        if not self.training:
            # in testing phase, we have Q == D
            assert doc_input_ids.size(0) == qry_input_ids.size(0), \
                'we expect same number of query/doc'
            tok_scores = self.compute_tok_score_pair(
                doc_reps, doc_input_ids,
                qry_reps, qry_input_ids, qry_attention_mask
            )

            # compute cls score separately
            cls_scores = (qry_cls * doc_cls).sum(-1)

            # sum the scores
            if self.model_args.no_cls:
                scores = tok_scores
            elif self.model_args.cls_only:
                scores = cls_scores
            else:
                if self.train_args.fp16:
                    with autocast(False):
                        scores = tok_scores.float() + cls_scores.float()  # B
                else:
                    scores = tok_scores + cls_scores  # B

            # loss not defined during inference
            return scores.view(-1)

        else:
            # for training phase, we have D = Q * group_size
            if self.model_args.x_device_negatives:
                # the idea is simple
                # fake it as if everything is on current device
                # gradient is taken care of at reduction time
                doc_input_ids, doc_cls, doc_reps = self.gather_tensors(
                    doc_input_ids, doc_cls, doc_reps)
                qry_input_ids, qry_attention_mask, qry_cls, qry_reps = self.gather_tensors(
                    qry_input_ids, qry_attention_mask, qry_cls, qry_reps)

            # qry_reps: Q * LQ * d
            # doc_reps: D * LD * d
            tok_scores = self.compute_tok_score_cart(
                doc_reps, doc_input_ids,
                qry_reps, qry_input_ids, qry_attention_mask
            )

            # remove padding and cls token
            if self.model_args.no_cls:
                scores = tok_scores
            elif self.model_args.cls_only:
                scores = torch.matmul(qry_cls, doc_cls.transpose(1, 0))  # Q * D
            else:
                cls_scores = torch.matmul(qry_cls, doc_cls.transpose(1, 0))  # Q * D
                with autocast(False):
                    scores = tok_scores.float() + cls_scores.float()  # Q * D

            labels = torch.arange(
                scores.size(0),
                device=doc_input['input_ids'].device,
                dtype=torch.long
            )
            # offset the labels
            labels = labels * self.data_args.train_group_size
            loss = self.cross_entropy(scores, labels)

            return loss, scores.view(-1)

    def mask_sep(self, qry_attention_mask):
        if self.model_args.no_sep:
            sep_pos = qry_attention_mask.sum(1).unsqueeze(1) - 1  # the sep token position
            _zeros = torch.zeros_like(sep_pos)
            qry_attention_mask.scatter_(1, sep_pos.long(), _zeros)

        return qry_attention_mask

    def compute_tok_score_pair(self, doc_reps, doc_input_ids, qry_reps, qry_input_ids, qry_attention_mask):
        exact_match = qry_input_ids.unsqueeze(2) == doc_input_ids.unsqueeze(1)  # B * LQ * LD
        exact_match = exact_match.float()
        # qry_reps: B * LQ * d
        # doc_reps: B * LD * d
        scores_no_masking = torch.bmm(qry_reps, doc_reps.permute(0, 2, 1))  # B * LQ * LD
        if self.model_args.pooling == 'max':
            tok_scores, _ = (scores_no_masking * exact_match).max(dim=2)  # B * LQ
        else:
            raise NotImplementedError(f'{self.model_args.pooling} pooling is not defined')
        # remove padding and cls token
        tok_scores = (tok_scores * qry_attention_mask)[:, 1:].sum(-1)
        return tok_scores

    def compute_tok_score_cart(self, doc_reps, doc_input_ids, qry_reps, qry_input_ids, qry_attention_mask):
        qry_input_ids = qry_input_ids.unsqueeze(2).unsqueeze(3)  # Q * LQ * 1 * 1
        doc_input_ids = doc_input_ids.unsqueeze(0).unsqueeze(1)  # 1 * 1 * D * LD
        exact_match = doc_input_ids == qry_input_ids  # Q * LQ * D * LD
        exact_match = exact_match.float()
        scores_no_masking = torch.matmul(
            qry_reps.view(-1, self.model_args.token_dim),  # (Q * LQ) * d
            doc_reps.view(-1, self.model_args.token_dim).transpose(0, 1)  # d * (D * LD)
        )
        scores_no_masking = scores_no_masking.view(
            *qry_reps.shape[:2], *doc_reps.shape[:2])  # Q * LQ * D * LD
        # scores_no_masking = scores_no_masking.permute(0, 2, 1, 3)  # Q * D * LQ * LD
        if self.model_args.pooling == 'max':
            scores, _ = (scores_no_masking * exact_match).max(dim=3)  # Q * LQ * D
        else:
            raise NotImplementedError(f'{self.model_args.pooling} pooling is not defined')

        tok_scores = (scores * qry_attention_mask.unsqueeze(2))[:, 1:].sum(1)
        return tok_scores

    def _gather_tensor(self, t: Tensor):
        all_tensors = [torch.empty_like(t) for _ in range(dist.get_world_size())]
        dist.all_gather(all_tensors, t)
        all_tensors[self.train_args.local_rank] = t
        return all_tensors

    def gather_tensors(self, *tt: Tensor):
        tt = [torch.cat(self._gather_tensor(t)) for t in tt]
        return tt


import string

STOPLIST = ["a", "about", "also", "am", "an", "and", "another", "any", "anyone", "are", "aren't", "as", "at", "be",
            "been", "being", "but", "by", "despite", "did", "didn't", "do", "does", "doesn't", "doing", "done", "don't",
            "each", "etc", "every", "everyone", "for", "from", "further", "had", "hadn't", "has", "hasn't", "have",
            "haven't", "having", "he", "he'd", "he'll", "her", "here", "here's", "hers", "herself", "he's",
            "him", "himself", "his", "however", "i", "i'd", "if", "i'll", "i'm", "in", "into", "is", "isn't", "it",
            "its", "it's", "itself", "i've", "just", "let's", "like", "lot", "may", "me", "might", "mightn't",
            "my", "myself", "no", "nor", "not", "of", "on", "onto", "or", "other", "ought", "oughtn't", "our", "ours",
            "ourselves", "out", "over", "shall", "shan't", "she", "she'd", "she'll", "she's", "since", "so", "some",
            "something", "such", "than", "that", "that's", "the", "their", "theirs", "them", "themselves", "then",
            "there", "there's", "these", "they", "they'd", "they'll", "they're", "they've", "this", "those", "through",
            "tht", "to", "too", "usually", "very", "via", "was", "wasn't", "we", "we'd", "well", "we'll", "were",
            "we're", "weren't", "we've", "will", "with", "without", "won't", "would", "wouldn't", "yes", "yet", "you",
            "you'd", "you'll", "your", "you're", "yours", "yourself", "yourselves", "you've"]

printable = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ ')
printableX = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-. ')
printable3X = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ,.- ')

printableD = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.- ')
printable3D = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.- ')

STOPLIST_ = list(map(lambda s: ''.join(filter(lambda x: x in printable, s)), STOPLIST))

STOPLIST = {}
for w in STOPLIST_:
    STOPLIST[w] = True

def clean_passage(s, join=True):
    s = [(x.lower() if x in printable3X else ' ') for x in s]
    s = [(x if x in printableX else ' ' + x + ' ') for x in s]
    s = ''.join(s).split()
    s = [(w if '.' not in w else (' . ' if len(max(w.split('.'), key=len)) > 1 else '').join(w.split('.'))) for w in s]
    s = ' '.join(s).split()
    s = [(w if '-' not in w else w.replace('-', '') + ' ( ' + ' '.join(w.split('-')) + ' ) ') for w in s]
    s = ' '.join(s).split()
    # s = [w for w in s if w not in STOPLIST]

    return ' '.join(s) if join else s


def clean_query(s, join=True):
    s = [(x.lower() if x in printable3D else ' ') for x in s]
    s = [(x if x in printableD else ' ' + x + ' ') for x in s]
    s = ''.join(s).split()
    s = [(w if '.' not in w else (' ' if len(max(w.split('.'), key=len)) > 1 else '').join(w.split('.'))) for w in s]
    s = ' '.join(s).split()
    s = [(w if '-' not in w else (' ' if len(min(w.split('-'), key=len)) > 1 else '').join(w.split('-'))) for w in s]
    s = ' '.join(s).split()
    s = [w for w in s if w not in STOPLIST]

    return ' '.join(s) if join else s


import torch
from transformers import AutoTokenizer, AutoModel

class SPARTA(torch.nn.Module):
    def __init__(self, model_name, device): #SpanBERT/spanbert-base-cased'): #bert-base-uncased    #distilbert-base-uncased #distilroberta-base
        super().__init__()
        self.bert_model = AutoModel.from_pretrained(model_name)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.bert_model.to(device)
        self.score_bias = torch.nn.Parameter(torch.tensor([1.0], device=device))
        self.device = device
        self.max_length = 300

    def bert_embeddings(self, input_ids):
        return self.bert_model.embeddings.word_embeddings(input_ids)

    def query_embeddings(self, query):
        queries_batch = self.tokenizer(query, padding=True, truncation=True, return_tensors='pt', add_special_tokens=False, max_length=self.max_length).to(self.device)
        queries_embeddings = self.bert_embeddings(queries_batch['input_ids'])
        return queries_embeddings

    def passage_embeddings(self, passages):
        passage_batch = self.tokenizer(passages, padding=True, truncation=True, return_tensors='pt', max_length=self.max_length).to(self.device)
        passage_embeddings = self.bert_model(**passage_batch).last_hidden_state
        return passage_embeddings

    def compute_scores(self, query_embeddings, passage_embeddings):
        ### Eq. 4 - Term matching
        scores = []
        for idx in range(len(query_embeddings)):        #TODO: use correct pytorch function for this
            scores.append(torch.matmul(query_embeddings[idx], passage_embeddings.transpose(1, 2)))
        scores = torch.stack(scores)
        #print("Scores:", scores.shape)
        max_scores = torch.max(scores, dim=-1).values
        #print("Max-Scores:", max_scores.shape)

        ### Eq. 5 - ReLu
        relu_scores = torch.relu(max_scores)    #torch.relu(max_scores + self.score_bias)  #Bias score does not change that much?
        #print("ReLu-Scores:", relu_scores.shape)

        ### Eq. 6 - Final Score
        final_scores = torch.sum(torch.log(relu_scores + 1), dim=-1) #.unsqueeze(dim=0)
        #print("Final scores:", final_scores.shape)
        return final_scores

    def forward(self, queries, passages):
        query_embeddings = self.query_embeddings(queries)
        passage_embeddings = self.passage_embeddings(passages)
        return self.compute_scores(query_embeddings, passage_embeddings)